import logging
from typing import Any, Callable, Dict, List, Optional, Tuple

from deepchem.data import Dataset
from deepchem.trans import Transformer
from deepchem.models import Model
from deepchem.metrics import Metric

logger = logging.getLogger(__name__)


def _convert_hyperparam_dict_to_filename(hyper_params: Dict[str, Any]) -> str:
    """Helper function that converts a dictionary of hyperparameters to a string that can be a filename.

    Parameters
    ----------
    hyper_params: Dict
        Maps string of hyperparameter name to int/float/string/list etc.

    Returns
    -------
    filename: str
        A filename of form "_key1_value1_value2_..._key2..."
    """
    filename = ""
    keys = sorted(hyper_params.keys())
    for key in keys:
        filename += "_%s" % str(key)
        value = hyper_params[key]
        if isinstance(value, int):
            filename += "_%s" % str(value)
        elif isinstance(value, float):
            filename += "_%f" % value
        else:
            filename += "%s" % str(value)
    return filename


class HyperparamOpt(object):
    """Abstract superclass for hyperparameter search classes.

    This class is an abstract base class for hyperparameter search
    classes in DeepChem. Hyperparameter search is performed on
    `dc.models.Model` classes. Each hyperparameter object accepts a
    `dc.models.Model` class upon construct. When the `hyperparam_search`
    class is invoked, this class is used to construct many different
    concrete models which are trained on the specified training set and
    evaluated on a given validation set.

    Different subclasses of `HyperparamOpt` differ in the choice of
    strategy for searching the hyperparameter evaluation space. This
    class itself is an abstract superclass and should never be directly
    instantiated.
    """

    def __init__(self, model_builder: Callable[..., Model]):
        """Initialize Hyperparameter Optimizer.

        Note this is an abstract constructor which should only be used by
        subclasses.

        Parameters
        ----------
        model_builder: constructor function.
            This parameter must be constructor function which returns an
            object which is an instance of `dc.models.Model`. This function
            must accept two arguments, `model_params` of type `dict` and
            `model_dir`, a string specifying a path to a model directory.
            See the example.
        """
        if self.__class__.__name__ == "HyperparamOpt":
            raise ValueError(
                "HyperparamOpt is an abstract superclass and cannot be directly instantiated. \
          You probably want to instantiate a concrete subclass instead.")
        self.model_builder = model_builder

    def hyperparam_search(
            self,
            params_dict: Dict,
            train_dataset: Dataset,
            valid_dataset: Dataset,
            metric: Metric,
            output_transformers: List[Transformer] = [],
            nb_epoch: int = 10,
            use_max: bool = True,
            logfile: str = 'results.txt',
            logdir: Optional[str] = None,
            **kwargs) -> Tuple[Model, Dict[str, Any], Dict[str, Any]]:
        """Conduct Hyperparameter search.

        This method defines the common API shared by all hyperparameter
        optimization subclasses. Different classes will implement
        different search methods but they must all follow this common API.

        Parameters
        ----------
        params_dict: Dict
            Dictionary mapping strings to values. Note that the
            precise semantics of `params_dict` will change depending on the
            optimizer that you're using. Depending on the type of
            hyperparameter optimization, these values can be
            ints/floats/strings/lists/etc. Read the documentation for the
            concrete hyperparameter optimization subclass you're using to
            learn more about what's expected.
        train_dataset: Dataset
            dataset used for training
        valid_dataset: Dataset
            dataset used for validation(optimization on valid scores)
        metric: Metric
            metric used for evaluation
        output_transformers: list[Transformer]
            Transformers for evaluation. This argument is needed since
            `train_dataset` and `valid_dataset` may have been transformed
            for learning and need the transform to be inverted before
            the metric can be evaluated on a model.
        nb_epoch: int, (default 10)
            Specifies the number of training epochs during each iteration of optimization.
        use_max: bool, optional
            If True, return the model with the highest score. Else return
            model with the minimum score.
        logdir: str, optional
            The directory in which to store created models. If not set, will
            use a temporary directory.
        logfile: str, optional (default `results.txt`)
            Name of logfile to write results to. If specified, this must
            be a valid file name. If not specified, results of hyperparameter
            search will be written to `logdir/results.txt`.

        Returns
        -------
        Tuple[`best_model`, `best_hyperparams`, `all_scores`]
            `(best_model, best_hyperparams, all_scores)` where `best_model` is
            an instance of `dc.models.Model`, `best_hyperparams` is a
            dictionary of parameters, and `all_scores` is a dictionary mapping
            string representations of hyperparameter sets to validation
            scores.
        """
        raise NotImplementedError



import random
import copy
import torch
import numpy as np
from torch_geometric.nn import GINEConv, global_add_pool, global_mean_pool, global_max_pool
from torch_geometric.nn.aggr import AttentionalAggregation, Set2Set
from torch_geometric.nn.inits import uniform
import torch.nn as nn
from torch.functional import F
from deepchem.data import Dataset
from deepchem.models.losses import SoftmaxCrossEntropy, EdgePredictionLoss, GraphNodeMaskingLoss, GraphEdgeMaskingLoss, DeepGraphInfomaxLoss, GraphContextPredLoss
from deepchem.models.torch_models import ModularTorchModel
from deepchem.feat.graph_data import BatchGraphData, GraphData, shortest_path_length
from typing import Iterable, List, Tuple
from deepchem.metrics import to_one_hot

num_node_type = 120  # including the extra mask tokens
num_chirality_tag = 3
# Relevant in future PRs
num_edge_type = 6  # including aromatic and self-loop edge, and extra masked tokens
# num_edge_direction = 3


class GNN(torch.nn.Module):
    """
    GNN module for the GNNModular model.

    This module is responsible for the graph neural network layers in the GNNModular model.

    Parameters
    ----------
    node_type_embedding: torch.nn.Embedding
        Embedding layer for node types.
    chirality_embedding: torch.nn.Embedding
        Embedding layer for chirality tags.
    gconvs: torch.nn.ModuleList
        ModuleList of graph convolutional layers.
    batch_norms: torch.nn.ModuleList
        ModuleList of batch normalization layers.
    dropout: int
        Dropout probability.
    jump_knowledge: str
        The type of jump knowledge to use. [1] Must be one of "last", "sum", "max", "concat" or "none".
        "last": Use the node representation from the last GNN layer.
        "concat": Concatenate the node representations from all GNN layers.
        "max": Take the element-wise maximum of the node representations from all GNN layers.
        "sum": Take the element-wise sum of the node representations from all GNN layers.
    init_emb: bool
        Whether to initialize the embedding layers with Xavier uniform initialization.

    References
    ----------
    .. [1] Xu, K. et al. Representation Learning on Graphs with Jumping Knowledge Networks. Preprint at https://doi.org/10.48550/arXiv.1806.03536 (2018).

    Example
    -------
    >>> from deepchem.models.torch_models.gnn import GNNModular
    >>> from deepchem.feat.graph_data import BatchGraphData
    >>> from deepchem.feat.molecule_featurizers import SNAPFeaturizer
    >>> featurizer = SNAPFeaturizer()
    >>> smiles = ["C1=CC=CC=C1", "C1=CC=CC=C1C=O", "C1=CC=CC=C1C(=O)O"]
    >>> features = featurizer.featurize(smiles)
    >>> modular = GNNModular(emb_dim = 8, task = "edge_pred")
    >>> batched_graph = BatchGraphData(features).numpy_to_torch(device=modular.device)
    >>> gnnmodel = modular.gnn
    >>> print(gnnmodel(batched_graph)[0].shape)
    torch.Size([23, 8])

    """

    def __init__(self,
                 node_type_embedding,
                 chirality_embedding,
                 gconvs,
                 batch_norms,
                 dropout,
                 jump_knowledge,
                 init_emb=False):
        super(GNN, self).__init__()

        self.node_type_embedding = node_type_embedding
        self.chirality_embedding = chirality_embedding
        self.gconv = gconvs
        self.batch_norms = batch_norms
        self.dropout = dropout
        self.num_layer = len(gconvs)
        self.jump_knowledge = jump_knowledge

        # may mess with loading pretrained weights
        if init_emb:
            torch.nn.init.xavier_uniform_(self.node_type_embedding.weight.data)
            torch.nn.init.xavier_uniform_(self.chirality_embedding.weight.data)

    def forward(self, data: BatchGraphData):
        """
        Forward pass for the GNN module.

        Parameters
        ----------
        data: BatchGraphData
            Batched graph data.
        """

        node_feats = data.node_features[:, 0].long()  # type: ignore
        chiral_feats = data.node_features[:, 1].long()  # type: ignore
        node_emb = self.node_type_embedding(node_feats)
        chir_emb = self.chirality_embedding(chiral_feats)
        x = node_emb + chir_emb

        h_list = [x]
        for i, conv_layer in enumerate(self.gconv):
            h = conv_layer(h_list[i], data.edge_index, data.edge_features)
            h = self.batch_norms[i](h)
            h = F.dropout(F.relu(h), self.dropout, training=self.training)
            if i == self.num_layer - 1:
                # remove relu for the last layer
                h = F.dropout(h, self.dropout, training=self.training)
            else:
                h = F.dropout(F.relu(h), self.dropout, training=self.training)
            h_list.append(h)

        # Different implementations of jump_knowledge
        if self.jump_knowledge == "concat":
            node_representation = torch.cat(h_list, dim=1)
            # reshapes node_representation to (num_nodes, num_layers * emb_dim)
        elif self.jump_knowledge == "last":
            node_representation = h_list[-1]
        elif self.jump_knowledge == "max":
            h_list = [h.unsqueeze_(0) for h in h_list]
            node_representation = torch.max(torch.cat(h_list, dim=0), dim=0)[0]
        elif self.jump_knowledge == "sum":
            h_list = [h.unsqueeze_(0) for h in h_list]
            node_representation = torch.sum(torch.cat(h_list, dim=0), dim=0)[0]

        return (node_representation, data)


class GNNHead(torch.nn.Module):
    """
    Prediction head module for the GNNModular model.

    Parameters
    ----------
    pool: Union[function,torch.nn.Module]
        Pooling function or nn.Module to use
    head: torch.nn.Module
        Prediction head to use
    task: str
        The type of task. Must be one of "regression", "classification".
    num_tasks: int
        Number of tasks.
    num_classes: int
        Number of classes for classification.
    """

    def __init__(self, pool, head, task, num_tasks, num_classes):
        super().__init__()
        self.pool = pool
        self.head = head
        self.task = task
        self.num_tasks = num_tasks
        self.num_classes = num_classes

    def forward(self, data):
        """
        Forward pass for the GNN head module.

        Parameters
        ----------
        data: tuple
            A tuple containing the node representations and the input graph data.
            node_representation is a torch.Tensor created after passing input through the GNN layers.
            input_batch is the original input BatchGraphData.
        """
        node_representation, input_batch = data

        pooled = self.pool(node_representation, input_batch.graph_index)
        out = self.head(pooled)
        if self.task == "classification":
            out = torch.reshape(out, (-1, self.num_tasks, self.num_classes))
        return out


class LocalGlobalDiscriminator(nn.Module):
    """
    This discriminator module is a linear layer without bias, used to measure the similarity between local node representations (`x`) and global graph representations (`summary`).

    The goal of the discriminator is to distinguish between positive and negative pairs of local and global representations.

    Examples
    --------
    >>> import torch
    >>> from deepchem.models.torch_models.gnn import LocalGlobalDiscriminator
    >>> discriminator = LocalGlobalDiscriminator(hidden_dim=64)
    >>> x = torch.randn(32, 64)  # Local node representations
    >>> summary = torch.randn(32, 64)  # Global graph representations
    >>> similarity_scores = discriminator(x, summary)
    >>> print(similarity_scores.shape)
    torch.Size([32])
    """

    def __init__(self, hidden_dim):
        """
        `self.weight` is a learnable weight matrix of shape `(hidden_dim, hidden_dim)`.

        nn.Parameters are tensors that require gradients and are optimized during the training process.

        Parameters
        ----------
        hidden_dim : int
            The size of the hidden dimension for the weight matrix.

        """
        super().__init__()
        self.weight = nn.Parameter(torch.Tensor(hidden_dim, hidden_dim))
        self.reset_parameters()

    def reset_parameters(self):
        size = self.weight.size(0)
        uniform(size, self.weight)

    def forward(self, x, summary):
        """
        Computes the product of `summary` and `self.weight`, and then calculates the element-wise product of `x` and the resulting matrix `h`.
        It then sums over the `hidden_dim` dimension, resulting in a tensor of shape `(batch_size,)`, which represents the similarity scores between the local and global representations.

        Parameters
        ----------
        x : torch.Tensor
            Local node representations of shape `(batch_size, hidden_dim)`.
        summary : torch.Tensor
            Global graph representations of shape `(batch_size, hidden_dim)`.

        Returns
        -------
        torch.Tensor
            A tensor of shape `(batch_size,)`, representing the similarity scores between the local and global representations.
        """
        h = torch.matmul(summary, self.weight)
        return torch.sum(x * h, dim=1)


class GNNModular(ModularTorchModel):
    """
    Modular GNN which allows for easy swapping of GNN layers.

    Parameters
    ----------
    gnn_type: str
        The type of GNN layer to use. Must be one of "gin", "gcn", "graphsage", or "gat".
    num_layer: int
        The number of GNN layers to use.
    emb_dim: int
        The dimensionality of the node embeddings.
    num_tasks: int
        The number of tasks.
    graph_pooling: str
        The type of graph pooling to use. Must be one of "sum", "mean", "max", "attention" or "set2set".
        "sum" may cause issues with positive prediction loss.
    dropout: float, optional (default 0)
        The dropout probability.
    jump_knowledge: str, optional (default "last")
        The type of jump knowledge to use. [1] Must be one of "last", "sum", "max", or "concat".
        "last": Use the node representation from the last GNN layer.
        "concat": Concatenate the node representations from all GNN layers. This will increase the dimensionality of the node representations by a factor of `num_layer`.
        "max": Take the element-wise maximum of the node representations from all GNN layers.
        "sum": Take the element-wise sum of the node representations from all GNN layers. This may cause issues with positive prediction loss.
    task: str, optional (default "regression")
        The type of task.
        Unsupervised tasks:
        edge_pred: Edge prediction. Predicts whether an edge exists between two nodes.
        mask_nodes: Masking nodes. Predicts the masked node.
        mask_edges: Masking edges. Predicts the masked edge.
        infomax: Infomax. Maximizes mutual information between local node representations and a pooled global graph representation.
        context_pred: Context prediction. Predicts the surrounding context of a node.
        Supervised tasks:
        "regression" or "classification".
    mask_rate: float, optional (default 0.1)
        The rate at which to mask nodes or edges for mask_nodes and mask_edges tasks.
    mask_edge: bool, optional (default True)
        Whether to also mask connecting edges for mask_nodes tasks.
    context_size: int, optional (default 1)
        The size of the context to use for context prediction tasks.
    neighborhood_size: int, optional (default 3)
        The size of the neighborhood to use for context prediction tasks.
    context_mode: str, optional (default "cbow")
        The context mode to use for context prediction tasks. Must be one of "cbow" or "skipgram".
    neg_samples: int, optional (default 1)
        The number of negative samples to use for context prediction.

    Examples
    --------
    >>> import numpy as np
    >>> import deepchem as dc
    >>> from deepchem.feat.molecule_featurizers import SNAPFeaturizer
    >>> from deepchem.models.torch_models.gnn import GNNModular
    >>> featurizer = SNAPFeaturizer()
    >>> smiles = ["C1=CC=CC=C1", "C1=CC=CC=C1C=O", "C1=CC=CC=C1C(=O)O"]
    >>> features = featurizer.featurize(smiles)
    >>> dataset = dc.data.NumpyDataset(features, np.zeros(len(features)))
    >>> model = GNNModular(task="edge_pred")
    >>> loss = model.fit(dataset, nb_epoch=1)

    References
    ----------
    .. [1] Xu, K. et al. Representation Learning on Graphs with Jumping Knowledge Networks. Preprint at https://doi.org/10.48550/arXiv.1806.03536 (2018).
    .. [2] Hu, W. et al. Strategies for Pre-training Graph Neural Networks. Preprint at https://doi.org/10.48550/arXiv.1905.12265 (2020).
    """

    def __init__(self,
                 gnn_type: str = "gin",
                 num_layer: int = 3,
                 emb_dim: int = 64,
                 num_tasks: int = 1,
                 num_classes: int = 2,
                 graph_pooling: str = "mean",
                 dropout: int = 0,
                 jump_knowledge: str = "last",
                 task: str = "edge_pred",
                 mask_rate: float = .1,
                 mask_edge: bool = True,
                 context_size: int = 1,
                 neighborhood_size: int = 3,
                 context_mode: str = "cbow",
                 neg_samples: int = 1,
                 **kwargs):
        self.gnn_type = gnn_type
        self.num_layer = num_layer
        self.emb_dim = emb_dim

        self.num_tasks = num_tasks
        self.num_classes = num_classes
        if task == "classification":
            self.output_dim = num_classes * num_tasks
            self.criterion = SoftmaxCrossEntropy()._create_pytorch_loss()
        elif task == "regression":
            self.output_dim = num_tasks
            self.criterion = F.mse_loss
        elif task == "edge_pred":
            self.output_dim = num_tasks
            self.edge_pred_loss = EdgePredictionLoss()._create_pytorch_loss()
        elif task == "mask_nodes":
            self.mask_rate = mask_rate
            self.mask_edge = mask_edge
            self.node_mask_loss = GraphNodeMaskingLoss()._create_pytorch_loss(
                self.mask_edge)
        elif task == "mask_edges":
            self.mask_rate = mask_rate
            self.edge_mask_loss = GraphEdgeMaskingLoss()._create_pytorch_loss()
        elif task == "infomax":
            self.infomax_loss = DeepGraphInfomaxLoss()._create_pytorch_loss()
        elif task == "context_pred":
            self.context_size = context_size
            self.neighborhood_size = neighborhood_size
            self.neg_samples = neg_samples
            self.context_mode = context_mode
            self.context_pred_loss = GraphContextPredLoss(
            )._create_pytorch_loss(context_mode, neg_samples)

        self.graph_pooling = graph_pooling
        self.dropout = dropout
        self.jump_knowledge = jump_knowledge
        self.task = task

        self.components = self.build_components()
        self.model = self.build_model()
        super().__init__(self.model, self.components, **kwargs)

    def build_components(self):
        """
        Builds the components of the GNNModular model. It initializes the encoders, batch normalization layers, pooling layers, and head layers based on the provided configuration. The method returns a dictionary containing the following components:

        Components list, type and description:
        --------------------------------------
        node_type_embedding: torch.nn.Embedding, an embedding layer for node types.

        chirality_embedding: torch.nn.Embedding, an embedding layer for chirality tags.

        gconvs: torch_geometric.nn.conv.MessagePassing, a list of graph convolutional layers (encoders) based on the specified GNN type (GIN, GCN, or GAT).

        batch_norms: torch.nn.BatchNorm1d, a list of batch normalization layers corresponding to the encoders.

        pool: Union[function,torch_geometric.nn.aggr.Aggregation], a pooling layer based on the specified graph pooling type (sum, mean, max, attention, or set2set).

        head: nn.Linear, a linear layer for the head of the model.

        These components are then used to construct the GNN and GNN_head modules for the GNNModular model.
        """

        encoders, batch_norms = self.build_gnn(self.num_layer)
        components = {
            'node_type_embedding':
                torch.nn.Embedding(num_node_type, self.emb_dim),
            'chirality_embedding':
                torch.nn.Embedding(num_chirality_tag, self.emb_dim),
            'gconvs':
                encoders,
            'batch_norms':
                batch_norms
        }
        self.gnn = GNN(components['node_type_embedding'],
                       components['chirality_embedding'], components['gconvs'],
                       components['batch_norms'], self.dropout,
                       self.jump_knowledge)

        if self.task in ("mask_nodes", "mask_edges"):
            linear_pred_nodes = torch.nn.Linear(self.emb_dim, num_node_type -
                                                1)  # -1 to remove mask token
            linear_pred_edges = torch.nn.Linear(self.emb_dim, num_edge_type -
                                                1)  # -1 to remove mask token
            components.update({
                'linear_pred_nodes': linear_pred_nodes,
                'linear_pred_edges': linear_pred_edges
            })

        # for supervised tasks, add prediction head
        elif self.task in ("regression", "classification"):
            if self.graph_pooling == "sum":
                pool = global_add_pool
            elif self.graph_pooling == "mean":
                pool = global_mean_pool
            elif self.graph_pooling == "max":
                pool = global_max_pool
            elif self.graph_pooling == "attention":
                if self.jump_knowledge == "concat":
                    pool = AttentionalAggregation(
                        gate_nn=torch.nn.Linear((self.num_layer + 1) *
                                                self.emb_dim, 1))
                else:
                    pool = AttentionalAggregation(
                        gate_nn=torch.nn.Linear(self.emb_dim, 1))
            elif self.graph_pooling == "set2set":
                set2setiter = 3
                if self.jump_knowledge == "concat":
                    pool = Set2Set((self.num_layer + 1) * self.emb_dim,
                                   set2setiter)
                else:
                    pool = Set2Set(self.emb_dim, processing_steps=set2setiter)

            if self.graph_pooling == "set2set":
                mult = 2
            else:
                mult = 1

            if self.jump_knowledge == "concat":
                head = torch.nn.Linear(
                    mult * (self.num_layer + 1) * self.emb_dim, self.output_dim)
            else:
                head = torch.nn.Linear(mult * self.emb_dim, self.output_dim)

            components.update({'pool': pool, 'head': head})

            self.gnn_head = GNNHead(components['pool'], components['head'],
                                    self.task, self.num_tasks, self.num_classes)

        elif self.task == 'infomax':
            descrim = LocalGlobalDiscriminator(self.emb_dim)
            components.update({
                'discriminator': descrim,
                'pool': global_mean_pool
            })

        elif self.task == 'context_pred':
            if self.graph_pooling == "sum":
                pool = global_add_pool
            elif self.graph_pooling == "mean":
                pool = global_mean_pool
            elif self.graph_pooling == "max":
                pool = global_max_pool
            elif self.graph_pooling == "attention":
                raise NotImplementedError(
                    "Attentional pooling not implemented for context prediction task."
                )
            elif self.graph_pooling == "set2set":
                raise NotImplementedError(
                    "Set2set pooling not implemented for context prediction task."
                )

            if self.jump_knowledge == "concat":  # concat changes the emb_dim
                c_gconvs, c_batch_norms = self.build_gnn(self.num_layer)
            else:
                c_gconvs, c_batch_norms = self.build_gnn(
                    self.neighborhood_size - self.context_size)
            context_gnn_components = {
                'c_node_type_embedding':
                    torch.nn.Embedding(num_node_type, self.emb_dim),
                'c_chirality_embedding':
                    torch.nn.Embedding(num_chirality_tag, self.emb_dim),
                'c_gconvs':
                    c_gconvs,
                'c_batch_norms':
                    c_batch_norms
            }

            self.context_gnn = GNN(
                context_gnn_components['c_node_type_embedding'],
                context_gnn_components['c_chirality_embedding'],
                context_gnn_components['c_gconvs'],
                context_gnn_components['c_batch_norms'], self.dropout,
                self.jump_knowledge)
            components.update({'pool': pool, **context_gnn_components})

        return components

    def build_gnn(self, num_layer):
        """
        Build graph neural network encoding layers by specifying the number of GNN layers.

        Parameters
        ----------
        num_layer : int
            The number of GNN layers to be created.

        Returns
        -------
        tuple of (torch.nn.ModuleList, torch.nn.ModuleList)
            A tuple containing two ModuleLists:
            1. encoders: A ModuleList of GNN layers (currently only GIN is supported).
            2. batch_norms: A ModuleList of batch normalization layers corresponding to each GNN layer.
        """

        encoders = []
        batch_norms = []
        for layer in range(num_layer):
            if self.gnn_type == "gin":
                encoders.append(
                    GINEConv(
                        torch.nn.Linear(self.emb_dim, self.emb_dim),
                        edge_dim=2,  # edge type, edge direction
                        aggr="add"))
            else:
                raise ValueError("Only GIN is supported for now")
            # Relevent for future PRs
            # elif self.gnn_type == "gcn":
            #     encoders.append(GCNConv(self.emb_dim))
            # elif self.gnn_type == "gat":
            #     encoders.append(GATConv(self.emb_dim))
            batch_norms.append(torch.nn.BatchNorm1d(self.emb_dim))
        encoders = torch.nn.ModuleList(encoders)
        batch_norms = torch.nn.ModuleList(batch_norms)

        return encoders, batch_norms

    def build_model(self):
        """
        Builds the appropriate model based on the specified task.

        For the edge prediction task, the model is simply the GNN module because it is an unsupervised task and does not require a prediction head.

        Supervised tasks such as node classification and graph regression require a prediction head, so the model is a sequential module consisting of the GNN module followed by the GNN_head module.
        """
        # unsupervised tasks do not need a pred head
        if self.task in ("edge_pred", "mask_nodes", "mask_edges", "infomax",
                         "context_pred"):
            return self.gnn
        elif self.task in ("regression", "classification"):
            return torch.nn.Sequential(self.gnn, self.gnn_head)
        else:
            raise ValueError(f"Task {self.task} is not supported.")

    def loss_func(self, inputs, labels, weights):
        """
        The loss function executed in the training loop, which is based on the specified task.
        """
        if self.task == "edge_pred":
            node_emb, inputs = self.model(inputs)
            loss = self.edge_pred_loss(node_emb, inputs)
        elif self.task == "mask_nodes":
            loss = self.masked_node_loss_loader(inputs)
        elif self.task == "mask_edges":
            loss = self.masked_edge_loss_loader(inputs)
        elif self.task == "infomax":
            loss = self.infomax_loss_loader(inputs)
        elif self.task == "regression":
            loss = self.regression_loss_loader(inputs, labels)
        elif self.task == "classification":
            loss = self.classification_loss_loader(inputs, labels)
        elif self.task == "context_pred":
            loss = self.context_pred_loss_loader(inputs)
        return (loss * weights).mean()

    def regression_loss_loader(self, inputs, labels):
        out = self.model(inputs)
        reg_loss = self.criterion(out, labels)
        return reg_loss

    def classification_loss_loader(self, inputs, labels):
        out = self.model(inputs)
        out = F.softmax(out, dim=2)
        class_loss = self.criterion(out, labels)
        return class_loss

    def masked_node_loss_loader(self, inputs):
        """
        Produces the loss between the predicted node features and the true node features for masked nodes.  Set mask_edge to True to also predict the edge types for masked edges.
        """

        node_emb, inputs = self.model(inputs)
        pred_node = self.components['linear_pred_nodes'](
            node_emb[inputs.masked_node_indices])
        if self.mask_edge:
            masked_edge_index = inputs.edge_index[:,
                                                  inputs.connected_edge_indices]
            edge_rep = node_emb[masked_edge_index[0]] + node_emb[
                masked_edge_index[1]]
            pred_edge = self.components['linear_pred_edges'](edge_rep)
        else:
            pred_edge = None
        return self.node_mask_loss(pred_node, pred_edge, inputs)

    def masked_edge_loss_loader(self, inputs):
        """
        Produces the loss between the predicted edge types and the true edge types for masked edges.
        """
        node_emb, inputs = self.model(inputs)
        # predict the edge types
        masked_edge_index = inputs.edge_index[:, inputs.masked_edge_idx]
        edge_emb = node_emb[masked_edge_index[0]] + node_emb[
            masked_edge_index[1]]
        pred_edge = self.components['linear_pred_edges'](edge_emb)

        return self.edge_mask_loss(pred_edge, inputs)

    def infomax_loss_loader(self, inputs):
        """
        Loss that maximizes mutual information between local node representations and a pooled global graph representation. The positive and negative scores represent the similarity between local node representations and global graph representations of simlar and dissimilar graphs, respectively.

        Parameters
        ----------
        inputs: BatchedGraphData
            BatchedGraphData object containing the node features, edge indices, and graph indices for the batch of graphs.
        """
        node_emb, inputs = self.model(inputs)
        summary_emb = torch.sigmoid(self.components['pool'](node_emb,
                                                            inputs.graph_index))
        positive_expanded_summary_emb = summary_emb[inputs.graph_index]

        shifted_summary_emb = summary_emb[cycle_index(len(summary_emb), 1)]
        negative_expanded_summary_emb = shifted_summary_emb[inputs.graph_index]

        positive_score = self.components['discriminator'](
            node_emb, positive_expanded_summary_emb)
        negative_score = self.components['discriminator'](
            node_emb, negative_expanded_summary_emb)

        return self.infomax_loss(positive_score, negative_score)

    def context_pred_loss_loader(self, inputs):
        """
        Loads the context prediction loss for the given input by taking the batched subgraph and context graphs and computing the context prediction loss for each subgraph and context graph pair.

        Parameters
        ----------
        inputs : tuple
            A tuple containing the following elements:
            - substruct_batch (BatchedGraphData): Batched subgraph, or neighborhood, graphs.
            - s_overlap (List[int]): List of overlapping subgraph node indices between the subgraph and context graphs.
            - context_graphs (BatchedGraphData): Batched context graphs.
            - c_overlap (List[int]): List of overlapping context node indices between the subgraph and context graphs.
            - overlap_size (List[int]): List of the number of overlapping nodes between the subgraph and context graphs.

        Returns
        -------
        context_pred_loss : torch.Tensor
            The context prediction loss
        """
        substruct_batch = inputs[0]
        s_overlap = inputs[1]
        context_graphs = inputs[2]
        c_overlap = inputs[3]
        overlap_size = inputs[4]

        substruct_rep = self.gnn(substruct_batch)[0][
            s_overlap]  # 0 for node representation index
        overlapped_node_rep = self.context_gnn(context_graphs)[0][
            c_overlap]  # 0 for node representation index

        context_rep = self.components['pool'](overlapped_node_rep, c_overlap)
        # negative contexts are obtained by shifting the indicies of context embeddings
        neg_context_rep = torch.cat([
            context_rep[cycle_index(len(context_rep), i + 1)]
            for i in range(self.neg_samples)
        ],
                                    dim=0)

        context_pred_loss = self.context_pred_loss(substruct_rep,
                                                   overlapped_node_rep,
                                                   context_rep, neg_context_rep,
                                                   overlap_size)

        return context_pred_loss

    def _overlap_batcher(self, substruct_graphs, s_overlap, context_graphs,
                         c_overlap):
        """
        This method provides batching for the context prediction task.

        It handles the batching of the overlapping indicies between the subgraph and context graphs.

        Parameters
        ----------
        substruct_graphs: BatchedGraphData
            Batched subgraph, or neighborhood, graphs.
        s_overlap: List[List[int]]
            List of lists of overlapping subgraph node indicies between the subgraph and context graphs.
        context_graphs: BatchedGraphData
            Batched context graphs.
        c_overlap: List[List[int]]
            List of lists of overlapping context node indicies between the subgraph and context graphs.

        Returns
        -------
        flat_s_overlap: List[int]
            List of overlapping subgraph node indicies between the subgraph and context graphs.
        flat_c_overlap: List[int]
            List of overlapping context node indicies between the subgraph and context graphs.
        overlap_size: List[int]
            List of the number of overlapping nodes between the subgraph and context graphs.
        """
        cumsum_substruct = 0
        cumsum_context = 0

        for i, (sub, context) in enumerate(zip(substruct_graphs,
                                               context_graphs)):
            num_nodes_substruct = len(sub.node_features)
            num_nodes_context = len(context.node_features)

            s_overlap[i] = [s + cumsum_substruct for s in s_overlap[i]]
            c_overlap[i] = [c + cumsum_context for c in c_overlap[i]]

            cumsum_substruct += num_nodes_substruct
            cumsum_context += num_nodes_context

        flat_s_overlap = [item for sublist in s_overlap for item in sublist]
        flat_c_overlap = [item for sublist in c_overlap for item in sublist]
        overlap_size = [len(s) for s in c_overlap]
        return flat_s_overlap, flat_c_overlap, overlap_size

    def _prepare_batch(self, batch):
        """
        Prepares the batch for the model by converting the GraphData numpy arrays to BatchedGraphData torch tensors and moving them to the device, then transforming the input to the appropriate format for the task.

        Parameters
        ----------
        batch: tuple
            A tuple containing the inputs, labels, and weights for the batch.

        Returns
        -------
        inputs: BatchGraphData
            The inputs for the batch, converted to a BatchGraphData object, moved to the device, and transformed to the appropriate format for the task.
        labels: torch.Tensor
            The labels for the batch, moved to the device.
        weights: torch.Tensor
            The weights for the batch, moved to the device.
        """
        inputs, labels, weights = batch
        if self.task in ("regression", "classification", "infomax"):
            inputs = BatchGraphData(inputs[0]).numpy_to_torch(self.device)
        if self.task == "edge_pred":
            inputs = BatchGraphData(inputs[0]).numpy_to_torch(self.device)
            inputs = negative_edge_sampler(inputs)
        elif self.task == "mask_nodes":
            inputs = BatchGraphData(inputs[0]).numpy_to_torch(self.device)
            inputs = mask_nodes(inputs, self.mask_rate)
        elif self.task == "mask_edges":
            inputs = BatchGraphData(inputs[0]).numpy_to_torch(self.device)
            inputs = mask_edges(inputs, self.mask_rate)
        elif self.task == "context_pred":
            sampled_g = [
                context_pred_sampler(graph, self.context_size,
                                     self.neighborhood_size)
                for graph in inputs[0]
            ]
            try:
                subgraphs_list = [x[0] for x in sampled_g]
                s_overlap_list = [x[1] for x in sampled_g]
                context_list = [x[2] for x in sampled_g]
                c_overlap_list = [x[3] for x in sampled_g]
            except ValueError:
                raise ValueError(
                    "Not enough nodes in graph to sample context, use a smaller context or a larger neighborhood size."
                )

            s_overlap, c_overlap, overlap_size = self._overlap_batcher(
                subgraphs_list, s_overlap_list, context_list, c_overlap_list)
            s_overlap = torch.tensor(s_overlap).to(self.device)
            c_overlap = torch.tensor(c_overlap).to(self.device)

            b_subgraphs = BatchGraphData(subgraphs_list).numpy_to_torch(
                self.device)
            b_context = BatchGraphData(context_list).numpy_to_torch(self.device)
            inputs = (b_subgraphs, s_overlap, b_context, c_overlap,
                      overlap_size)

        _, labels, weights = super()._prepare_batch(([], labels, weights))

        if (len(labels) != 0) and (len(weights) != 0):
            labels = labels[0]
            weights = weights[0]

        return inputs, labels, weights

    def default_generator(
            self,
            dataset: Dataset,
            epochs: int = 1,
            mode: str = 'fit',
            deterministic: bool = True,
            pad_batches: bool = True) -> Iterable[Tuple[List, List, List]]:
        """
        This default generator is modified from the default generator in dc.models.tensorgraph.tensor_graph.py to support multitask classification. If the task is classification, the labels y_b are converted to a one-hot encoding and reshaped according to the number of tasks and classes.
        """

        for epoch in range(epochs):
            for (X_b, y_b, w_b,
                 ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                               deterministic=deterministic,
                                               pad_batches=pad_batches):
                if self.task == 'classification' and y_b is not None:
                    y_b = to_one_hot(y_b.flatten(), self.num_classes).reshape(
                        -1, self.num_tasks, self.num_classes)
                yield ([X_b], [y_b], [w_b])


def negative_edge_sampler(input_graph: BatchGraphData):
    """
    NegativeEdge is a function that adds negative edges to the input graph data. It randomly samples negative edges (edges that do not exist in the original graph) and adds them to the input graph data.
    The number of negative edges added is equal to half the number of edges in the original graph. This is useful for tasks like edge prediction, where the model needs to learn to differentiate between existing and non-existing edges.

    Parameters
    ----------
    input_graph: dc.feat.graph_data.BatchGraphData
        The input graph data.

    Returns
    -------
    BatchGraphData
        A new BatchGraphData object with the additional attribute `negative_edge_index`.

    Example
    -------
    >>> import numpy as np
    >>> import torch
    >>> from deepchem.feat.graph_data import BatchGraphData, GraphData
    >>> from deepchem.models.torch_models.gnn import negative_edge_sampler
    >>> num_nodes_list, num_edge_list = [3, 4, 5], [2, 4, 5]
    >>> num_node_features, num_edge_features = 32, 32
    >>> edge_index_list = [
    ...     np.array([[0, 1], [1, 2]]),
    ...     np.array([[0, 1, 2, 3], [1, 2, 0, 2]]),
    ...     np.array([[0, 1, 2, 3, 4], [1, 2, 3, 4, 0]]),
    ... ]
    >>> graph_list = [
    ...     GraphData(node_features=np.random.random_sample(
    ...         (num_nodes_list[i], num_node_features)),
    ...               edge_index=edge_index_list[i],
    ...               edge_features=np.random.random_sample(
    ...                   (num_edge_list[i], num_edge_features)),
    ...               node_pos_features=None) for i in range(len(num_edge_list))
    ... ]
    >>> batched_graph = BatchGraphData(graph_list)
    >>> batched_graph = batched_graph.numpy_to_torch()
    >>> neg_sampled = negative_edge_sampler(batched_graph)
    """
    data = copy.deepcopy(input_graph)

    num_nodes = data.num_nodes
    num_edges = data.num_edges

    edge_set = set([
        str(data.edge_index[0, i].cpu().item()) + "," +
        str(data.edge_index[1, i].cpu().item())
        for i in range(data.edge_index.shape[1])
    ])

    redandunt_sample = torch.randint(0, num_nodes, (2, 5 * num_edges))
    sampled_ind = []
    sampled_edge_set = set([])
    for i in range(5 * num_edges):
        node1 = redandunt_sample[0, i].cpu().item()
        node2 = redandunt_sample[1, i].cpu().item()
        edge_str = str(node1) + "," + str(node2)
        if edge_str not in edge_set and edge_str not in sampled_edge_set and not node1 == node2:
            sampled_edge_set.add(edge_str)
            sampled_ind.append(i)
        if len(sampled_ind) == num_edges / 2:
            break

    data.negative_edge_index = redandunt_sample[:, sampled_ind]  # type: ignore

    return data


def mask_nodes(input_graph: BatchGraphData,
               mask_rate,
               masked_node_indices=None,
               mask_edge=True):
    """
    Mask nodes and their connected edges in a BatchGraphData object.

    This function assumes that the first node feature is the atomic number, for example with the SNAPFeaturizer. It will set masked nodes' features to 0.

    Parameters
    ----------
    input_graph: dc.feat.BatchGraphData
        Assume that the edge ordering is the default PyTorch geometric ordering, where the two directions of a single edge occur in pairs.
        Eg. data.edge_index = tensor([[0, 1, 1, 2, 2, 3],
                                    [1, 0, 2, 1, 3, 2]])
    masked_node_indices: list, optional
        If None, then randomly samples num_nodes * mask rate number of node indices. Otherwise, a list of node indices that sets the nodes to be masked (for debugging only).
    mask_edge: bool, optional
        Will mask the edges connected to the masked nodes.

    Returns
    -------
    data: dc.feat.BatchGraphData
        Creates new attributes in the original data object:
        - data.mask_node_idx
        - data.mask_node_label
        - data.mask_edge_idx
        - data.mask_edge_label

        """
    data = copy.deepcopy(input_graph)

    if masked_node_indices is None:
        # sample x distinct nodes to be masked, based on mask rate. But
        # will sample at least 1 node
        num_nodes = data.node_features.size()[0]  # type: ignore
        sample_size = int(num_nodes * mask_rate + 1)
        masked_node_indices = random.sample(range(num_nodes), sample_size)

    # create mask node label by copying node feature of mask node
    mask_node_labels_list = []
    for node_idx in masked_node_indices:
        mask_node_labels_list.append(data.node_features[node_idx].view(1, -1))
    data.mask_node_label = torch.cat(  # type: ignore
        mask_node_labels_list, dim=0)[:, 0].long()
    data.masked_node_indices = torch.tensor(masked_node_indices)  # type: ignore

    # modify the original node feature of the masked node
    num_node_feats = data.node_features.size()[1]  # type: ignore
    for node_idx in masked_node_indices:
        data.node_features[node_idx] = torch.zeros((1, num_node_feats))
    # zeros are meant to represent the masked features. This is distinct from the
    # original implementation, where the masked features are represented by the
    # the last feature token 119.
    # link to source: https://github.com/snap-stanford/pretrain-gnns/blob/08f126ac13623e551a396dd5e511d766f9d4f8ff/chem/util.py#L241

    if mask_edge:
        # create mask edge labels by copying edge features of edges that are connected to
        # mask nodes
        connected_edge_indices = []
        for edge_idx, (u, v) in enumerate(
                data.edge_index.cpu().numpy().T):  # type: ignore
            for node_idx in masked_node_indices:
                if node_idx in set(
                    (u, v)) and edge_idx not in connected_edge_indices:
                    connected_edge_indices.append(edge_idx)

        if len(connected_edge_indices) > 0:
            # create mask edge labels by copying edge features of the edges connected to
            # the mask nodes
            mask_edge_labels_list = []
            for edge_idx in connected_edge_indices[::2]:  # because the
                # edge ordering is such that two directions of a single
                # edge occur in pairs, so to get the unique undirected
                # edge indices, we take every 2nd edge index from list
                mask_edge_labels_list.append(
                    data.edge_features[edge_idx].view(  # type: ignore
                        1, -1))

            data.mask_edge_label = torch.cat(  # type: ignore
                mask_edge_labels_list, dim=0)[:, 0].long()  # type: ignore
            # modify the original edge features of the edges connected to the mask nodes
            num_edge_feat = data.edge_features.size()[1]  # type: ignore
            for edge_idx in connected_edge_indices:
                data.edge_features[edge_idx] = torch.zeros(  # type: ignore
                    (1, num_edge_feat))  # type: ignore
            # zeros are meant to represent the masked features. This is distinct from the
            # original implementation, where the masked features are represented by the
            # the last feature token 4.
            # link to source: https://github.com/snap-stanford/pretrain-gnns/blob/08f126ac13623e551a396dd5e511d766f9d4f8ff/chem/util.py#L268

            data.connected_edge_indices = torch.tensor(  # type: ignore
                connected_edge_indices[::2])
        else:
            data.mask_edge_label = torch.empty(  # type: ignore
                (0, 2)).to(torch.int64)
            data.connected_edge_indices = torch.tensor(  # type: ignore
                connected_edge_indices).to(torch.int64)

    return data


def mask_edges(input_graph: BatchGraphData,
               mask_rate: float,
               masked_edge_indices=None):
    """
    Mask edges in a BatchGraphData object.

    This is separate from the mask_nodes function because we want to be able to mask edges without masking any nodes.

    Parameters
    ----------
    input_graph: dc.feat.BatchGraphData
        Assume that the edge ordering is the default PyTorch geometric ordering, where the two directions of a single edge occur in pairs.
        Eg. data.edge_index = tensor([[0, 1, 1, 2, 2, 3],
                                      [1, 0, 2, 1, 3, 2]])
    masked_edge_indices : list, optional
        If None, then randomly sample num_edges * mask_rate + 1 number of edge indices. Otherwise should correspond to the 1st direction of an edge pair. ie all indices should be an even number

    Returns
    -------
    data: dc.feat.BatchGraphData
        Creates new attributes in the original  object:
        - data.mask_edge_idx: indices of masked edges
        - data.mask_edge_labels: corresponding ground truth edge feature for each masked edge
        - data.edge_attr: modified in place: the edge features (both directions) that correspond to the masked edges have the masked edge feature
        """
    data = copy.deepcopy(input_graph)

    if masked_edge_indices is None:
        # sample x distinct edges to be masked, based on mask rate. But
        # will sample at least 1 edge
        num_edges = int(data.edge_index.size()[1] /  # type: ignore
                        2)  # num unique edges
        sample_size = int(num_edges * mask_rate + 1)
        # during sampling, we only pick the 1st direction of a particular
        # edge pair
        masked_edge_indices = [
            2 * i for i in random.sample(range(num_edges), sample_size)
        ]

    data.masked_edge_idx = torch.tensor(  # type: ignore
        np.array(masked_edge_indices))

    # create ground truth edge features for the edges that correspond to
    # the masked indices
    mask_edge_labels_list = []
    for idx in masked_edge_indices:
        mask_edge_labels_list.append(  # yapf: disable
            data.edge_features[idx].view(  # type: ignore
                1, -1))
    data.mask_edge_label = torch.cat(  # type: ignore
        mask_edge_labels_list, dim=0)

    # created new masked edge_attr, where both directions of the masked
    # edges have masked edge type. For message passing in gcn

    # append the 2nd direction of the masked edges
    all_masked_edge_indices = masked_edge_indices + [
        i + 1 for i in masked_edge_indices
    ]
    num_edge_feat = data.edge_features.size()[1]  # type: ignore
    for idx in all_masked_edge_indices:
        data.edge_features[idx] = torch.zeros(  # type: ignore
            (1, num_edge_feat))
    # zeros are meant to represent the masked features. This is distinct from the
    # original implementation, where the masked features are represented by 0s and
    # an additional mask feature
    # link to source: https://github.com/snap-stanford/pretrain-gnns/blob/08f126ac13623e551a396dd5e511d766f9d4f8ff/bio/util.py#L101

    return data


def cycle_index(num, shift):
    """
    Creates a 1-dimensional tensor of integers with a specified length (`num`) and a cyclic shift (`shift`). The tensor starts with integers from `shift` to `num - 1`, and then wraps around to include integers from `0` to `shift - 1` at the end.

    Parameters
    ----------
    num: int
        Length of the tensor.
    shift: int
        Amount to shift the tensor by.

    Example
    -------
    >>> num = 10
    >>> shift = 3
    >>> arr = cycle_index(num, shift)
    >>> print(arr)
    tensor([3, 4, 5, 6, 7, 8, 9, 0, 1, 2])
    """
    arr = torch.arange(num) + shift
    arr[-shift:] = torch.arange(shift)
    return arr


def context_pred_sampler(input_graph: GraphData,
                         l1: int,
                         l2: int,
                         root_idx=None):
    """
    Generate subgraph and context graph for context prediction.

    This function takes an input graph and generates a subgraph and a context graph for context prediction. The subgraph is the entire input graph, while the context graph is a ring around the root node outside of l1 and inside of l2.

    Parameters
    ----------
    input_graph : GraphData
        The input graph for which the subgraph and context graph are to be generated.
    l1 : int
        The inner diameter of the context graph.
    l2 : int
        The outer diameter of the context graph.
    root_idx : int, optional
        The index of the root node. If not provided, a random node will be selected as the root.

    Returns
    -------
    subgraph : GraphData
        The subgraph generated from the input graph.
    s_overlap : list
        The indices of overlapping nodes between substruct and context, with respect to the substruct ordering.
    context_G : GraphData
        The context graph generated from the input graph.
    c_overlap : list
        The indices of overlapping nodes between substruct and context, with respect to the context ordering.

    Examples
    --------
    >>> import numpy as np
    >>> from deepchem.feat.graph_data import GraphData
    >>> from deepchem.models.torch_models.gnn import context_pred_sampler
    >>> x = np.array([[1, 0], [0, 1], [1, 1], [0, 0], [1, 0], [0, 1], [1, 1], [0, 0]])
    >>> edge_index = np.array([[0, 1, 2, 3, 4, 5, 6, 7], [1, 0, 3, 2, 5, 4, 7, 6]])
    >>> edge_feats = np.array([[1, 0], [0, 1], [1, 1], [0, 0], [1, 0], [0, 1], [1, 1], [0, 0]])
    >>> graph_index = np.array([0, 0, 1, 1, 2, 2, 3, 3])
    >>> data = GraphData(node_features=x,
    ...                  edge_index=edge_index,
    ...                  edge_features=edge_feats,
    ...                  graph_index=graph_index)
    >>> l1 = 1
    >>> l2 = 2
    >>> subgraph, s_overlap, context_G, c_overlap = context_pred_sampler(data, l1, l2)
    """
    data = copy.deepcopy(input_graph)

    root_idx = random.sample(range(data.num_nodes), 1)[0]

    # Take the entire graph, but can be modified to take a subgraph of k-hops from the root node
    x_substruct = data.node_features
    edge_attr_substruct = data.edge_features
    edge_index_substruct = data.edge_index

    subgraph = GraphData(node_features=x_substruct,
                         edge_features=edge_attr_substruct,
                         edge_index=edge_index_substruct)

    # Get node idx between root and the inner diameter l1
    l1_node_idxes = shortest_path_length(data, root_idx, l1).keys()
    # Get node idx between root and the outer diameter l2
    l2_node_idxes = shortest_path_length(data, root_idx, l2).keys()

    # takes a ring around the root node outside of l1 and inside of l2
    context_node_idxes = set(l1_node_idxes).symmetric_difference(
        set(l2_node_idxes))

    if len(context_node_idxes) > 0:
        context_G, context_node_map = data.subgraph(context_node_idxes)
        # Get indices of overlapping nodes between substruct and context, WRT context ordering
        s_overlap = list(context_node_idxes)
        c_overlap = [context_node_map[old_idx] for old_idx in s_overlap]
        return (subgraph, s_overlap, context_G, c_overlap)
    else:
        return (subgraph, [], None, [])



"""
Featurizers, transformers, and splitters for MolNet.
"""

import importlib
import inspect
import logging
from typing import Optional, Dict, Any

from deepchem.feat.base_classes import Featurizer
from deepchem.trans.transformers import Transformer
from deepchem.splits.splitters import Splitter

logger = logging.getLogger(__name__)


def get_defaults(module_name: Optional[str] = None) -> Dict[str, Any]:
    """Get featurizers, transformers, and splitters.

    This function returns a dictionary with class names as keys and classes
    as values. All MolNet ``load_x`` functions should specify which
    featurizers, transformers, and splitters the dataset supports and
    provide sensible defaults.

    Parameters
    ----------
    module_name : {"feat", "trans", "splits"}
        Default classes from deepchem.`module_name` will be returned.

    Returns
    -------
    defaults : Dict[str, Any]
        Keys are class names and values are class constructors.

    Examples
    --------
    >> splitter = get_defaults('splits')['RandomSplitter']()
    >> transformer = get_defaults('trans')['BalancingTransformer'](dataset, {"transform_X": True})
    >> featurizer = get_defaults('feat')["CoulombMatrix"](max_atoms=5)

    """

    if module_name not in ["feat", "trans", "splits"]:
        raise ValueError(
            "Input argument must be either 'feat', 'trans', or 'splits'.")

    if module_name == "feat":
        sc: Any = Featurizer
    elif module_name == "trans":
        sc = Transformer
    elif module_name == "splits":
        sc = Splitter

    module_name = "deepchem." + module_name

    module = importlib.import_module(module_name, package="deepchem")

    defaults = {
        x[0]: x[1]
        for x in inspect.getmembers(module, inspect.isclass)
        if issubclass(x[1], sc)
    }

    return defaults



import unittest
import os
import tempfile
from deepchem.utils import rdkit_utils
from deepchem.utils import pdbqt_utils


class TestPDBQTUtils(unittest.TestCase):

    def setUp(self):
        current_dir = os.path.dirname(os.path.realpath(__file__))
        self.protein_file = os.path.join(current_dir,
                                         "../../dock/tests/1jld_protein.pdb")
        self.ligand_file = os.path.join(current_dir,
                                        "../../dock/tests/1jld_ligand.sdf")

    def test_pdbqt_to_pdb(self):
        """Test that a PDBQT molecule can be converted back in to PDB."""
        xyz, mol = rdkit_utils.load_molecule(self.protein_file,
                                             calc_charges=False,
                                             add_hydrogens=False)
        with tempfile.TemporaryDirectory() as tmp:
            out_pdb = os.path.join(tmp, "mol.pdb")
            out_pdbqt = os.path.join(tmp, "mol.pdbqt")

            rdkit_utils.write_molecule(mol, out_pdb, is_protein=True)
            rdkit_utils.write_molecule(mol, out_pdbqt, is_protein=True)

            pdb_block = pdbqt_utils.pdbqt_to_pdb(out_pdbqt)
            from rdkit import Chem
            pdb_mol = Chem.MolFromPDBBlock(pdb_block,
                                           sanitize=False,
                                           removeHs=False)

            xyz, pdbqt_mol = rdkit_utils.load_molecule(out_pdbqt,
                                                       add_hydrogens=False,
                                                       calc_charges=False)

        assert pdb_mol.GetNumAtoms() == pdbqt_mol.GetNumAtoms()
        for atom_idx in range(pdb_mol.GetNumAtoms()):
            atom1 = pdb_mol.GetAtoms()[atom_idx]
            atom2 = pdbqt_mol.GetAtoms()[atom_idx]
            assert atom1.GetAtomicNum() == atom2.GetAtomicNum()

    def test_convert_mol_to_pdbqt(self):
        """Test that a ligand molecule can be coverted to PDBQT."""
        from rdkit import Chem
        xyz, mol = rdkit_utils.load_molecule(self.ligand_file,
                                             calc_charges=False,
                                             add_hydrogens=False)
        with tempfile.TemporaryDirectory() as tmp:
            outfile = os.path.join(tmp, "mol.pdbqt")
            writer = Chem.PDBWriter(outfile)
            writer.write(mol)
            writer.close()
            pdbqt_utils.convert_mol_to_pdbqt(mol, outfile)
            pdbqt_xyz, pdbqt_mol = rdkit_utils.load_molecule(
                outfile, add_hydrogens=False, calc_charges=False)
        assert pdbqt_mol.GetNumAtoms() == pdbqt_mol.GetNumAtoms()
        for atom_idx in range(pdbqt_mol.GetNumAtoms()):
            atom1 = pdbqt_mol.GetAtoms()[atom_idx]
            atom2 = pdbqt_mol.GetAtoms()[atom_idx]
            assert atom1.GetAtomicNum() == atom2.GetAtomicNum()

    def test_convert_protein_to_pdbqt(self):
        """Test a protein in a PDB can be converted to PDBQT."""
        from rdkit import Chem
        xyz, mol = rdkit_utils.load_molecule(self.protein_file,
                                             calc_charges=False,
                                             add_hydrogens=False)
        with tempfile.TemporaryDirectory() as tmp:
            outfile = os.path.join(tmp, "mol.pdbqt")
            writer = Chem.PDBWriter(outfile)
            writer.write(mol)
            writer.close()
            pdbqt_utils.convert_protein_to_pdbqt(mol, outfile)
            pdbqt_xyz, pdbqt_mol = rdkit_utils.load_molecule(
                outfile, add_hydrogens=False, calc_charges=False)
        assert pdbqt_mol.GetNumAtoms() == pdbqt_mol.GetNumAtoms()
        for atom_idx in range(pdbqt_mol.GetNumAtoms()):
            atom1 = pdbqt_mol.GetAtoms()[atom_idx]
            atom2 = pdbqt_mol.GetAtoms()[atom_idx]
            assert atom1.GetAtomicNum() == atom2.GetAtomicNum()



"""
Common code for loading MoleculeNet datasets.
"""
import os
import logging
import deepchem as dc
from deepchem.data import Dataset, DiskDataset
from typing import List, Optional, Tuple, Type, Union

logger = logging.getLogger(__name__)


class TransformerGenerator(object):
    """Create Transformers for Datasets.

    When loading molnet datasets, you cannot directly pass in Transformers
    to use because many Transformers require the Dataset they will be applied to
    as a constructor argument.  Instead you pass in TransformerGenerator objects
    which can create the Transformers once the Dataset is loaded.
    """

    def __init__(self, transformer_class: Type[dc.trans.Transformer], **kwargs):
        """Construct an object for creating Transformers.

        Parameters
        ----------
        transformer_class: Type[Transformer]
            the class of Transformer to create
        kwargs:
            any additional arguments are passed to the Transformer's constructor
        """
        self.transformer_class = transformer_class
        self.kwargs = kwargs

    def create_transformer(self, dataset: Dataset) -> dc.trans.Transformer:
        """Construct a Transformer for a Dataset."""
        return self.transformer_class(dataset=dataset, **self.kwargs)

    def get_directory_name(self) -> str:
        """Get a name for directories on disk describing this Transformer."""
        name = self.transformer_class.__name__
        for key, value in self.kwargs.items():
            if isinstance(value, list):
                continue
            name += '_' + key + '_' + str(value)
        return name


featurizers = {
    'ecfp': dc.feat.CircularFingerprint(size=1024),
    'graphconv': dc.feat.ConvMolFeaturizer(),
    'raw': dc.feat.RawFeaturizer(),
    'onehot': dc.feat.OneHotFeaturizer(),
    'smiles2img': dc.feat.SmilesToImage(img_size=80, img_spec='std'),
    'weave': dc.feat.WeaveFeaturizer(),
}

splitters = {
    'index': dc.splits.IndexSplitter(),
    'random': dc.splits.RandomSplitter(),
    'scaffold': dc.splits.ScaffoldSplitter(),
    'butina': dc.splits.ButinaSplitter(),
    'fingerprint': dc.splits.FingerprintSplitter(),
    'task': dc.splits.TaskSplitter(),
    'stratified': dc.splits.RandomStratifiedSplitter()
}

transformers = {
    'balancing':
        TransformerGenerator(dc.trans.BalancingTransformer),
    'normalization':
        TransformerGenerator(dc.trans.NormalizationTransformer,
                             transform_y=True),
    'minmax':
        TransformerGenerator(dc.trans.MinMaxTransformer, transform_y=True),
    'clipping':
        TransformerGenerator(dc.trans.ClippingTransformer, transform_y=True),
    'log':
        TransformerGenerator(dc.trans.LogTransformer, transform_y=True)
}


class _MolnetLoader(object):
    """The class provides common functionality used by many molnet loader functions.
       It is an abstract class.  Subclasses implement loading of particular datasets.
    """

    def __init__(self, featurizer: Union[dc.feat.Featurizer, str],
                 splitter: Union[dc.splits.Splitter, str, None],
                 transformer_generators: List[Union[TransformerGenerator,
                                                    str]], tasks: List[str],
                 data_dir: Optional[str], save_dir: Optional[str], **kwargs):
        """Construct an object for loading a dataset.

        Parameters
        ----------
        featurizer: Featurizer or str
            the featurizer to use for processing the data.  Alternatively you can pass
            one of the names from dc.molnet.featurizers as a shortcut.
        splitter: Splitter or str
            the splitter to use for splitting the data into training, validation, and
            test sets.  Alternatively you can pass one of the names from
            dc.molnet.splitters as a shortcut.  If this is None, all the data
            will be included in a single dataset.
        transformer_generators: list of TransformerGenerators or strings
            the Transformers to apply to the data.  Each one is specified by a
            TransformerGenerator or, as a shortcut, one of the names from
            dc.molnet.transformers.
        tasks: List[str]
            the names of the tasks in the dataset
        data_dir: str
            a directory to save the raw data in
        save_dir: str
            a directory to save the dataset in
        """
        if 'split' in kwargs:
            splitter = kwargs['split']
            logger.warning("'split' is deprecated.  Use 'splitter' instead.")
        if isinstance(featurizer, str):
            featurizer = featurizers[featurizer.lower()]
        if isinstance(splitter, str):
            splitter = splitters[splitter.lower()]
        if data_dir is None:
            data_dir = dc.utils.data_utils.get_data_dir()
        if save_dir is None:
            save_dir = dc.utils.data_utils.get_data_dir()
        self.featurizer = featurizer
        self.splitter = splitter
        self.transformers = [
            transformers[t.lower()] if isinstance(t, str) else t
            for t in transformer_generators
        ]
        self.tasks = list(tasks)
        self.data_dir = data_dir
        self.save_dir = save_dir
        self.args = kwargs

    def load_dataset(
        self, name: str, reload: bool
    ) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
        """Load the dataset.

        Parameters
        ----------
        name: str
            the name of the dataset, used to identify the directory on disk
        reload: bool
            if True, the first call for a particular featurizer and splitter will cache
            the datasets to disk, and subsequent calls will reload the cached datasets.
        """
        # Build the path to the dataset on disk.

        featurizer_name = str(self.featurizer)
        splitter_name = 'None' if self.splitter is None else str(self.splitter)
        save_folder = os.path.join(self.save_dir, name + "-featurized",
                                   featurizer_name, splitter_name)
        if len(self.transformers) > 0:
            transformer_name = '_'.join(
                t.get_directory_name() for t in self.transformers)
            save_folder = os.path.join(save_folder, transformer_name)

        # Try to reload cached datasets.

        if reload:
            if self.splitter is None:
                if os.path.exists(save_folder):
                    transformers = dc.utils.data_utils.load_transformers(
                        save_folder)
                    return self.tasks, (DiskDataset(save_folder),), transformers
            else:
                loaded, all_dataset, transformers = dc.utils.data_utils.load_dataset_from_disk(
                    save_folder)
                if all_dataset is not None:
                    return self.tasks, all_dataset, transformers

        # Create the dataset

        logger.info("About to featurize %s dataset." % name)
        dataset = self.create_dataset()

        # Split and transform the dataset.

        if self.splitter is None:
            transformer_dataset: Dataset = dataset
        else:
            logger.info("About to split dataset with {} splitter.".format(
                self.splitter.__class__.__name__))
            train, valid, test = self.splitter.train_valid_test_split(dataset)
            transformer_dataset = train
        transformers = [
            t.create_transformer(transformer_dataset) for t in self.transformers
        ]
        logger.info("About to transform data.")
        if self.splitter is None:
            for transformer in transformers:
                dataset = transformer.transform(dataset)
            if reload and isinstance(dataset, DiskDataset):
                dataset.move(save_folder)
                dc.utils.data_utils.save_transformers(save_folder, transformers)
            return self.tasks, (dataset,), transformers

        for transformer in transformers:
            train = transformer.transform(train)
            valid = transformer.transform(valid)
            test = transformer.transform(test)
        if reload and isinstance(train, DiskDataset) and isinstance(
                valid, DiskDataset) and isinstance(test, DiskDataset):
            dc.utils.data_utils.save_dataset_to_disk(save_folder, train, valid,
                                                     test, transformers)
        return self.tasks, (train, valid, test), transformers

    def create_dataset(self) -> Dataset:
        """Subclasses must implement this to load the dataset."""
        raise NotImplementedError()



import math
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from typing import List, Sequence, Optional, Any, Tuple
from rdkit import Chem
from deepchem.feat.graph_data import BatchGraphData
from deepchem.models.torch_models.modular import ModularTorchModel
from deepchem.models.torch_models.grover_layers import (
    GroverEmbedding, GroverBondVocabPredictor, GroverAtomVocabPredictor,
    GroverFunctionalGroupPredictor)
from deepchem.models.torch_models.readout import GroverReadout
from deepchem.feat.vocabulary_builders import GroverAtomVocabularyBuilder, GroverBondVocabularyBuilder
from deepchem.utils.grover import extract_grover_attributes


class GroverPretrain(nn.Module):
    """The Grover Pretrain module.

    The GroverPretrain module is used for training an embedding based on the Grover Pretraining task.
    Grover pretraining is a self-supervised task where an embedding is trained to learn the contextual
    information of atoms and bonds along with graph-level properties, which are functional groups
    in case of molecular graphs.

    Parameters
    ----------
    embedding: nn.Module
        An embedding layer to generate embedding from input molecular graph
    atom_vocab_task_atom: nn.Module
        A layer used for predicting atom vocabulary from atom features generated via atom hidden states.
    atom_vocab_task_bond: nn.Module
        A layer used for predicting atom vocabulary from atom features generated via bond hidden states.
    bond_vocab_task_atom: nn.Module
        A layer used for predicting bond vocabulary from bond features generated via atom hidden states.
    bond_vocab_task_bond: nn.Module
        A layer used for predicting bond vocabulary from bond features generated via bond hidden states.

    Returns
    -------
    prediction_logits: Tuple
        A tuple of prediction logits containing prediction logits of atom vocabulary task from atom hidden state,
    prediction logits for atom vocabulary task from bond hidden states, prediction logits for bond vocabulary task
    from atom hidden states, prediction logits for bond vocabulary task from bond hidden states, functional
    group prediction logits from atom embedding generated from atom and bond hidden states, functional group
    prediction logits from bond embedding generated from atom and bond hidden states.

    Example
    -------
    >>> import deepchem as dc
    >>> from deepchem.feat.graph_data import BatchGraphData
    >>> from deepchem.utils.grover import extract_grover_attributes
    >>> from deepchem.models.torch_models.grover import GroverPretrain
    >>> from deepchem.models.torch_models.grover_layers import GroverEmbedding, GroverAtomVocabPredictor, GroverBondVocabPredictor, GroverFunctionalGroupPredictor
    >>> smiles = ['CC', 'CCC', 'CC(=O)C']

    >>> fg = dc.feat.CircularFingerprint()
    >>> featurizer = dc.feat.GroverFeaturizer(features_generator=fg)

    >>> graphs = featurizer.featurize(smiles)
    >>> batched_graph = BatchGraphData(graphs)
    >>> grover_graph_attributes = extract_grover_attributes(batched_graph)
    >>> f_atoms, f_bonds, a2b, b2a, b2revb, a2a, a_scope, b_scope, _, _ = grover_graph_attributes
    >>> components = {}
    >>> components['embedding'] = GroverEmbedding(node_fdim=f_atoms.shape[1], edge_fdim=f_bonds.shape[1])
    >>> components['atom_vocab_task_atom'] = GroverAtomVocabPredictor(vocab_size=10, in_features=128)
    >>> components['atom_vocab_task_bond'] = GroverAtomVocabPredictor(vocab_size=10, in_features=128)
    >>> components['bond_vocab_task_atom'] = GroverBondVocabPredictor(vocab_size=10, in_features=128)
    >>> components['bond_vocab_task_bond'] = GroverBondVocabPredictor(vocab_size=10, in_features=128)
    >>> components['functional_group_predictor'] = GroverFunctionalGroupPredictor(10)
    >>> model = GroverPretrain(**components)

    >>> inputs = f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope, a2a
    >>> output = model(inputs)

    Reference
    ---------
    .. Rong, Yu, et al. "Self-supervised graph transformer on large-scale molecular data." Advances in Neural Information Processing Systems 33 (2020): 12559-12571.
    """

    def __init__(self, embedding: nn.Module, atom_vocab_task_atom: nn.Module,
                 atom_vocab_task_bond: nn.Module,
                 bond_vocab_task_atom: nn.Module,
                 bond_vocab_task_bond: nn.Module,
                 functional_group_predictor: nn.Module):
        super(GroverPretrain, self).__init__()
        self.embedding = embedding
        self.atom_vocab_task_atom = atom_vocab_task_atom
        self.atom_vocab_task_bond = atom_vocab_task_bond
        self.bond_vocab_task_atom = bond_vocab_task_atom
        self.bond_vocab_task_bond = bond_vocab_task_bond
        self.functional_group_predictor = functional_group_predictor

    def forward(self, graph_batch):
        """Forward function

        Parameters
        ----------
        graph_batch: List[torch.Tensor]
            A list containing grover graph attributes
        """
        _, _, _, _, _, atom_scope, bond_scope, _ = graph_batch
        atom_scope = atom_scope.data.cpu().numpy().tolist()
        bond_scope = bond_scope.data.cpu().numpy().tolist()

        embeddings = self.embedding(graph_batch)
        av_task_atom_pred = self.atom_vocab_task_atom(
            embeddings["atom_from_atom"])
        av_task_bond_pred = self.atom_vocab_task_bond(
            embeddings["atom_from_bond"])

        bv_task_atom_pred = self.bond_vocab_task_atom(
            embeddings["bond_from_atom"])
        bv_task_bond_pred = self.bond_vocab_task_bond(
            embeddings["bond_from_bond"])

        fg_prediction = self.functional_group_predictor(embeddings, atom_scope,
                                                        bond_scope)

        return av_task_atom_pred, av_task_bond_pred, bv_task_atom_pred, bv_task_bond_pred, fg_prediction[
            'atom_from_atom'], fg_prediction['atom_from_bond'], fg_prediction[
                'bond_from_atom'], fg_prediction['bond_from_bond']


class GroverFinetune(nn.Module):
    """Grover Finetune model.

    For a graph level prediction task, the GroverFinetune model uses node/edge embeddings
    output by the GroverEmbeddong layer and applies a readout function on it to get
    graph embeddings and use additional MLP layers to predict the property of the molecular graph.

    Parameters
    ----------
    embedding: nn.Module
        An embedding layer to generate embedding from input molecular graph
    readout: nn.Module
        A readout layer to perform readout atom and bond hidden states
    mol_atom_from_atom_ffn: nn.Module
        A feed forward network which learns representation from atom messages generated via atom hidden states of a molecular graph
    mol_atom_from_bond_ffn: nn.Module
        A feed forward network which learns representation from atom messages generated via bond hidden states of a molecular graph
    mode: str
        classification or regression

    Returns
    -------
    prediction_logits: torch.Tensor
        prediction logits

    Example
    -------
    >>> import deepchem as dc
    >>> from deepchem.feat.graph_data import BatchGraphData
    >>> from deepchem.utils.grover import extract_grover_attributes
    >>> from deepchem.models.torch_models.grover_layers import GroverEmbedding
    >>> from deepchem.models.torch_models.readout import GroverReadout
    >>> from deepchem.models.torch_models.grover import GroverFinetune
    >>> smiles = ['CC', 'CCC', 'CC(=O)C']
    >>> fg = dc.feat.CircularFingerprint()
    >>> featurizer = dc.feat.GroverFeaturizer(features_generator=fg)
    >>> graphs = featurizer.featurize(smiles)
    >>> batched_graph = BatchGraphData(graphs)
    >>> attributes = extract_grover_attributes(batched_graph)
    >>> components = {}
    >>> f_atoms, f_bonds, a2b, b2a, b2revb, a2a, a_scope, b_scope, fg_labels, additional_features = attributes
    >>> inputs = f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope, a2a
    >>> components = {}
    >>> components['embedding'] = GroverEmbedding(node_fdim=f_atoms.shape[1], edge_fdim=f_bonds.shape[1])
    >>> components['readout'] = GroverReadout(rtype="mean", in_features=128)
    >>> components['mol_atom_from_atom_ffn'] = nn.Linear(in_features=additional_features.shape[1]+ 128, out_features=128)
    >>> components['mol_atom_from_bond_ffn'] = nn.Linear(in_features=additional_features.shape[1] + 128, out_features=128)
    >>> model = GroverFinetune(**components, mode='regression', hidden_size=128)
    >>> model.training = False
    >>> output = model((inputs, additional_features))

    Reference
    ---------
    .. Rong, Yu, et al. "Self-supervised graph transformer on large-scale molecular data." Advances in Neural Information Processing Systems 33 (2020): 12559-12571.
    """

    def __init__(self,
                 embedding: nn.Module,
                 readout: nn.Module,
                 mol_atom_from_atom_ffn: nn.Module,
                 mol_atom_from_bond_ffn: nn.Module,
                 hidden_size: int = 128,
                 mode: str = 'regression',
                 n_tasks: int = 1,
                 n_classes: Optional[int] = None):
        super().__init__()
        self.embedding = embedding
        self.readout = readout
        self.mol_atom_from_atom_ffn = mol_atom_from_atom_ffn
        self.mol_atom_from_bond_ffn = mol_atom_from_bond_ffn
        self.n_tasks = n_tasks
        self.n_classes = n_classes
        self.mode = mode
        # the hidden size here is the output size of last layer in mol_atom_from_atom_ffn and mol_atom_from_bond_ffn components.
        # it is necessary that aforementioned components produces output tensor of same size.
        if self.mode == 'classification':
            assert n_classes is not None
            self.linear = nn.Linear(hidden_size,
                                    out_features=n_tasks * n_classes)
        elif self.mode == 'regression':
            self.linear = nn.Linear(hidden_size, out_features=n_tasks)

    def forward(self, inputs):
        """
        Parameters
        ----------
        inputs: Tuple
            grover batch graph attributes
        """
        graphbatch, additional_features = inputs
        _, _, _, _, _, atom_scope, bond_scope, _ = graphbatch
        output = self.embedding(graphbatch)

        mol_atom_from_bond_output = self.readout(output["atom_from_bond"],
                                                 atom_scope)
        mol_atom_from_atom_output = self.readout(output["atom_from_atom"],
                                                 atom_scope)

        if additional_features[0] is not None:
            additional_features = torch.from_numpy(
                np.stack(additional_features)).float()
            additional_features.to(output["atom_from_bond"])
            if len(additional_features.shape) == 1:
                additional_features = additional_features.view(
                    1, additional_features.shape[0])
            mol_atom_from_atom_output = torch.cat(
                [mol_atom_from_atom_output, additional_features], 1)
            mol_atom_from_bond_output = torch.cat(
                [mol_atom_from_bond_output, additional_features], 1)

        atom_ffn_output = self.mol_atom_from_atom_ffn(mol_atom_from_atom_output)
        bond_ffn_output = self.mol_atom_from_bond_ffn(mol_atom_from_bond_output)
        if self.training:
            # In training mode, we return atom level aggregated output and bond level aggregated output.
            # The training has an additional objective which ensures that atom and bond level aggregated outputs
            # are similar to each other apart from the objective of making the aggregated output closer to each
            # other.
            return atom_ffn_output, bond_ffn_output
        else:
            if self.mode == 'classification':
                atom_ffn_output = torch.sigmoid(atom_ffn_output)
                bond_ffn_output = torch.sigmoid(bond_ffn_output)
            output = (atom_ffn_output + bond_ffn_output) / 2
            output = self.linear(output)

            if self.mode == 'classification':
                if self.n_tasks == 1:
                    logits = output.view(-1, self.n_classes)
                    softmax_dim = 1
                else:
                    logits = output.view(-1, self.n_tasks, self.n_classes)
                    softmax_dim = 2
                proba = F.softmax(logits, dim=softmax_dim)
                return proba, logits
            elif self.mode == 'regression':
                return output


class GroverModel(ModularTorchModel):
    """GROVER model

    The GROVER model employs a self-supervised message passing transformer architecutre
    for learning molecular representation. The pretraining task can learn rich structural
    and semantic information of molecules from unlabelled molecular data, which can
    be leveraged by finetuning for downstream applications. To this end, GROVER integrates message
    passing networks into a transformer style architecture.

    Parameters
    ----------
    node_fdim: int
        the dimension of additional feature for node/atom.
    edge_fdim: int
        the dimension of additional feature for edge/bond.
    atom_vocab: GroverAtomVocabularyBuilder
        Grover atom vocabulary builder required during pretraining.
    bond_vocab: GroverBondVocabularyBuilder
        Grover bond vocabulary builder required during pretraining.
    hidden_size: int
        Size of hidden layers
    features_only: bool
        Uses only additional features in the feed-forward network, no graph network
    self_attention: bool, default False
        When set to True, a self-attention layer is used during graph readout operation.
    functional_group_size: int (default: 85)
        Size of functional group used in grover.
    features_dim: int
        Size of additional molecular features, like fingerprints.
    ffn_num_layers: int (default: 1)
        Number of linear layers to use for feature extraction from embeddings
    task: str (pretraining or finetuning)
        Pretraining or finetuning tasks.
    mode: str (classification or regression)
        Training mode (used only for finetuning)
    n_tasks: int, optional (default: 1)
        Number of tasks
    n_classes: int, optiona (default: 2)
        Number of target classes in classification mode
    model_dir: str
        Directory to save model checkpoints
    dropout: float, optional (default: 0.2)
        dropout value
    actionvation: str, optional (default: 'relu')
        supported activation function

    Example
    -------
    >>> import deepchem as dc
    >>> from deepchem.models.torch_models.grover import GroverModel
    >>> from deepchem.feat.vocabulary_builders import (GroverAtomVocabularyBuilder, GroverBondVocabularyBuilder)
    >>> import pandas as pd
    >>> import os
    >>> import tempfile
    >>> tmpdir = tempfile.mkdtemp()
    >>> df = pd.DataFrame({'smiles': ['CC', 'CCC'], 'preds': [0, 0]})
    >>> filepath = os.path.join(tmpdir, 'example.csv')
    >>> df.to_csv(filepath, index=False)
    >>> dataset_path = os.path.join(filepath)
    >>> loader = dc.data.CSVLoader(tasks=['preds'], featurizer=dc.feat.DummyFeaturizer(), feature_field=['smiles'])
    >>> dataset = loader.create_dataset(filepath)
    >>> av = GroverAtomVocabularyBuilder()
    >>> av.build(dataset)
    >>> bv = GroverBondVocabularyBuilder()
    >>> bv.build(dataset)
    >>> fg = dc.feat.CircularFingerprint()
    >>> loader2 = dc.data.CSVLoader(tasks=['preds'], featurizer=dc.feat.GroverFeaturizer(features_generator=fg), feature_field='smiles')
    >>> graph_data = loader2.create_dataset(filepath)
    >>> model = GroverModel(node_fdim=151, edge_fdim=165, atom_vocab=av, bond_vocab=bv, features_dim=2048, hidden_size=128, functional_group_size=85, mode='regression', task='finetuning', model_dir='gm')
    >>> loss = model.fit(graph_data, nb_epoch=1)

    Reference
    ---------
    .. Rong, Yu, et al. "Self-supervised graph transformer on large-scale molecular data." Advances in Neural Information Processing Systems 33 (2020): 12559-12571.
    """

    def __init__(self,
                 node_fdim: int,
                 edge_fdim: int,
                 atom_vocab: GroverAtomVocabularyBuilder,
                 bond_vocab: GroverBondVocabularyBuilder,
                 hidden_size: int,
                 self_attention=False,
                 features_only=False,
                 functional_group_size: int = 85,
                 features_dim=128,
                 dropout=0.2,
                 activation='relu',
                 task='pretraining',
                 ffn_num_layers=1,
                 mode: Optional[str] = None,
                 model_dir=None,
                 n_tasks: int = 1,
                 n_classes: Optional[int] = None,
                 **kwargs):
        assert task in ['pretraining', 'finetuning']
        self.ffn_num_layers = ffn_num_layers
        self.activation = activation
        self.node_fdim = node_fdim
        self.edge_fdim = edge_fdim
        self.atom_vocab = atom_vocab
        self.bond_vocab = bond_vocab
        self.atom_vocab_size = atom_vocab.size
        self.bond_vocab_size = bond_vocab.size
        self.task = task
        self.model_dir = model_dir
        self.hidden_size = hidden_size
        self.attn_hidden_size = hidden_size
        self.attn_out_size = hidden_size
        self.functional_group_size = functional_group_size
        self.self_attention = self_attention
        self.features_only = features_only
        self.features_dim = features_dim
        self.dropout = dropout
        self.mode = mode
        self.n_tasks = n_tasks
        self.n_classes = n_classes
        self.components = self.build_components()
        self.model = self.build_model()
        super().__init__(self.model,
                         self.components,
                         model_dir=self.model_dir,
                         **kwargs)
        # FIXME In the above step, we initialize modular torch model but
        # something is missing here. The attribute loss from TorchModel gets assigned `loss_func`
        # by super class initialization in ModularTorchModel but here we reinitialize it.
        self.loss = self.get_loss_func()

    def build_components(self):
        """Builds components for grover pretraining and finetuning model.

        .. list-table:: Components of pretraining model
           :widths: 25 25 50
           :header-rows: 1

           * - Component name
             - Type
             - Description
           * - `embedding`
             - Graph message passing network
             - A layer which accepts a molecular graph and produces an embedding for grover pretraining task
           * - `atom_vocab_task_atom`
             - Feed forward layer
             - A layer which accepts an embedding generated from atom hidden states and predicts atom vocabulary for grover pretraining task
           * - `atom_vocab_task_bond`
             - Feed forward layer
             - A layer which accepts an embedding generated from bond hidden states and predicts atom vocabulary for grover pretraining task
           * - `bond_vocab_task_atom`
             - Feed forward layer
             - A layer which accepts an embedding generated from atom hidden states and predicts bond vocabulary for grover pretraining task
           * - `bond_vocab_task_bond`
             - Feed forward layer
             - A layer which accepts an embedding generated from bond hidden states and predicts bond vocabulary for grover pretraining task
           * - `functional_group_predictor`
             - Feed forward layer
             - A layer which accepts an embedding generated from a graph readout and predicts functional group for grover pretraining task

        .. list-table:: Components of finetuning model

           * - Component name
             - Type
             - Description
           * - `embedding`
             - Graph message passing network
             - An embedding layer to generate embedding from input molecular graph
           * - `readout`
             - Feed forward layer
             - A readout layer to perform readout atom and bond hidden states
           * - `mol_atom_from_atom_ffn`
             - Feed forward layer
             - A feed forward network which learns representation from atom messages generated via atom hidden states of a molecular graph
           * - `mol_atom_from_bond_ffn`
             - Feed forward layer
             - A feed forward network which learns representation from atom messages generated via bond hidden states of a molecular graph
        """
        if self.task == 'pretraining':
            components = self._get_pretraining_components()
        elif self.task == 'finetuning':
            components = self._get_finetuning_components()
        return components

    def build_model(self):
        """Builds grover pretrain or finetune model based on task"""
        if self.task == 'pretraining':
            return GroverPretrain(**self.components)
        elif self.task == 'finetuning':
            return GroverFinetune(**self.components,
                                  mode=self.mode,
                                  hidden_size=self.hidden_size,
                                  n_tasks=self.n_tasks,
                                  n_classes=self.n_classes)

    def get_loss_func(self):
        """Returns loss function based on task"""
        if self.task == 'pretraining':
            from deepchem.models.losses import GroverPretrainLoss
            return GroverPretrainLoss()._create_pytorch_loss()
        elif self.task == 'finetuning':
            return self._finetuning_loss

    def loss_func(self, inputs, labels, weights):
        """Returns loss function which performs forward iteration based on task type"""
        if self.task == 'pretraining':
            return self._pretraining_loss(inputs, labels, weights)
        elif self.task == 'finetuning':
            return self._finetuning_loss(inputs, labels, weights)

    def _get_pretraining_components(self):
        """Return pretraining components.

        The component names are described in GroverModel.build_components method.
        """
        components = {}
        components['embedding'] = GroverEmbedding(node_fdim=self.node_fdim,
                                                  edge_fdim=self.edge_fdim)
        components['atom_vocab_task_atom'] = GroverAtomVocabPredictor(
            self.atom_vocab_size, self.hidden_size)
        components['atom_vocab_task_bond'] = GroverAtomVocabPredictor(
            self.atom_vocab_size, self.hidden_size)
        components['bond_vocab_task_atom'] = GroverBondVocabPredictor(
            self.bond_vocab_size, self.hidden_size)
        components['bond_vocab_task_bond'] = GroverBondVocabPredictor(
            self.bond_vocab_size, self.hidden_size)
        components[
            'functional_group_predictor'] = GroverFunctionalGroupPredictor(
                self.functional_group_size)
        return components

    def _get_finetuning_components(self):
        """Return finetuning components.

        The component names are described in GroverModel.build_components method.
        """
        components = {}
        components['embedding'] = GroverEmbedding(node_fdim=self.node_fdim,
                                                  edge_fdim=self.edge_fdim)
        if self.self_attention:
            components['readout'] = GroverReadout(
                rtype="self_attention",
                in_features=self.hidden_size,
                attn_hidden=self.attn_hidden_size,
                attn_out=self.attn_out_size)
        else:
            components['readout'] = GroverReadout(rtype="mean",
                                                  in_features=self.hidden_size)
        components['mol_atom_from_atom_ffn'] = self._create_ffn()
        components['mol_atom_from_bond_ffn'] = self._create_ffn()

        return components

    def _prepare_batch(self, batch):
        """Prepare batch method for preprating batch of data for finetuning and pretraining tasks"""
        if self.task == 'pretraining':
            return self._prepare_batch_for_pretraining(batch)
        elif self.task == 'finetuning':
            return self._prepare_batch_for_finetuning(batch)

    def _prepare_batch_for_pretraining(self, batch: Tuple[Any, Any, Any]):
        """Prepare batch for pretraining

        This method is used for batching a sequence of graph data objects using BatchGraphData method.
        It batches a graph and extracts attributes from the batched graph and return it as inputs for
        the model. It also performs generates labels for atom vocab prediction and bonc vocab
        prediction tasks in grover.

        Parameters
        ----------
        batch: Tuple[Sequence[GraphData], Any, Any]
            A batch of data containing grover molecular graphs, target prediction values and weights for datapoint.

        Returns
        -------
        inputs: Tuple
            Inputs for grover pretraining model
        labels: Dict[str, torch.Tensor]
            Labels for grover pretraining self-supervised task
        w: Any
            Weights of data point
        """
        X, y, w = batch
        batchgraph = BatchGraphData(X[0])
        fgroup_label = getattr(batchgraph, 'fg_labels')
        smiles_batch = getattr(batchgraph, 'smiles').reshape(-1).tolist()

        f_atoms, f_bonds, a2b, b2a, b2revb, a2a, a_scope, b_scope, _, _ = extract_grover_attributes(
            batchgraph)

        atom_vocab_label = torch.Tensor(
            self.atom_vocab_random_mask(self.atom_vocab, smiles_batch)).long()
        bond_vocab_label = torch.Tensor(
            self.bond_vocab_random_mask(self.bond_vocab, smiles_batch)).long()
        labels = {
            "av_task": atom_vocab_label,
            "bv_task": bond_vocab_label,
            "fg_task": torch.Tensor(fgroup_label)
        }
        inputs = (f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope, a2a)
        return inputs, labels, w

    def _prepare_batch_for_finetuning(self, batch: Tuple[Any, Any, Any]):
        """Prepare batch for finetuning task

        The method batches a sequence of grover graph data objects using BatchGraphData utility
        and extracts attributes from the batched graph. The extracted attributes are fed to
        the grover model as inputs.

        Parameters
        ----------
        batch: Tuple[Sequence[GraphData], Any, Any]
            A batch of data containing grover molecular graphs, target prediction values and weights for datapoint.

        Returns
        -------
        inputs: Tuple
            Inputs for grover finetuning model
        labels: Dict[str, torch.Tensor]
            Labels for grover finetuning task
        w: Any
            Weights of data point
        """
        X, y, w = batch
        batchgraph = BatchGraphData(X[0])
        if y is not None:
            labels = torch.FloatTensor(y[0])
        else:
            labels = None
        f_atoms, f_bonds, a2b, b2a, b2revb, a2a, a_scope, b_scope, _, additional_features = extract_grover_attributes(
            batchgraph)
        inputs = (f_atoms, f_bonds, a2b, b2a, b2revb, a_scope, b_scope,
                  a2a), additional_features
        return inputs, labels, w

    def _pretraining_loss(self,
                          inputs,
                          labels,
                          weights: Optional[List[Sequence]] = None,
                          dist_coff: float = 0.1):
        """Grover pretraining loss

        The Grover pretraining loss function performs a forward iteration and returns
        the loss value.

        Parameters
        ----------
        inputs: Tuple[torch.Tensor]
            extracted grover graph attributed
        labels: Dict[str, torch.Tensor]
            Target predictions
        weights: List[Sequence]
            Weight to assign to each datapoint
        dist_coff: float, default: 0.1
            Loss term weight for weighting closeness between embedding generated from atom hidden state and bond hidden state in atom vocabulary and bond vocabulary prediction tasks.

        Returns
        -------
        loss: torch.Tensor
            loss value
        """
        _, _, _, _, _, atom_scope, bond_scope, _ = inputs
        av_task_atom_pred, av_task_bond_pred, bv_task_atom_pred, bv_task_bond_pred, fg_prediction_atom_from_atom, fg_prediction_atom_from_bond, fg_prediction_bond_from_atom, fg_prediction_bond_from_bond = self.model(
            inputs)

        loss = self.loss(av_task_atom_pred,
                         av_task_bond_pred,
                         bv_task_atom_pred,
                         bv_task_bond_pred,
                         fg_prediction_atom_from_atom,
                         fg_prediction_atom_from_bond,
                         fg_prediction_bond_from_atom,
                         fg_prediction_bond_from_bond,
                         labels['av_task'],
                         labels['bv_task'],
                         labels['fg_task'],
                         weights=weights,
                         dist_coff=dist_coff)  # type: ignore
        return loss

    def _finetuning_loss(self, inputs, labels, weights, dist_coff=0.1):
        """Loss function for finetuning task

        The finetuning loss is a binary cross entropy loss for classification mode and
        mean squared error loss for regression mode. During training of the model, apart from
        learning the data distribution, the loss function is also used to make the embedding
        generated from atom hidden state and bond hidden state close to each other.

        Parameters
        ----------
        inputs: Tuple[torch.Tensor]
            extracted grover graph attributed
        labels: Dict[str, torch.Tensor]
            Target predictions
        weights: List[Sequence]
            Weight to assign to each datapoint
        dist_coff: float, default: 0.1
            Loss term weight for weighting closeness between embedding generated from atom hidden state and bond hidden state

        Returns
        -------
        loss: torch.Tensor
            loss value
        """
        if self.mode == 'classification':
            pred_loss = nn.BCEWithLogitsLoss()
        elif self.mode == 'regression':
            pred_loss = nn.MSELoss()

        preds = self.model(inputs)

        if not self.model.training:
            # in eval mode.
            return pred_loss(preds, labels)
        elif self.model.training:
            dist_loss = nn.MSELoss()
            dist = dist_loss(preds[0], preds[1])
            pred_loss1 = pred_loss(preds[0].mean(axis=-1).unsqueeze(-1), labels)
            pred_loss2 = pred_loss(preds[1].mean(axis=-1).unsqueeze(-1), labels)
            return pred_loss1 + pred_loss2 + dist_coff * dist

    def _create_ffn(self):
        """Creates feed-forward network for the finetune task"""
        if self.features_only:
            first_linear_dim = self.features_size + self.features_dim
        else:
            if self.self_attention:
                first_linear_dim = self.hidden_size * self.attn_out_size
                # Also adding features, this is optional
                first_linear_dim += self.features_dim
            else:
                first_linear_dim = self.hidden_size + self.features_dim
        dropout = nn.Dropout(self.dropout)

        if self.activation == 'relu':
            activation = nn.ReLU()

        ffn = [dropout, nn.Linear(first_linear_dim, self.hidden_size)]
        for i in range(self.ffn_num_layers - 1):
            ffn.extend([
                activation, dropout,
                nn.Linear(self.hidden_size, self.hidden_size)
            ])

        return nn.Sequential(*ffn)

    @staticmethod
    def atom_vocab_random_mask(atom_vocab: GroverAtomVocabularyBuilder,
                               smiles: List[str]) -> List[int]:
        """Random masking of atom labels from vocabulary

        For every atom in the list of SMILES string, the algorithm fetches the atoms
        context (vocab label) from the vocabulary provided and returns the vocabulary
        labels with a random masking (probability of masking = 0.15).

        Parameters
        ----------
        atom_vocab: GroverAtomVocabularyBuilder
            atom vocabulary
        smiles: List[str]
            a list of smiles string

        Returns
        -------
        vocab_label: List[int]
            atom vocab label with random masking

        Example
        -------
        >>> import deepchem as dc
        >>> from deepchem.models.torch_models.grover import GroverModel
        >>> from deepchem.feat.vocabulary_builders import GroverAtomVocabularyBuilder
        >>> smiles = np.array(['CC', 'CCC'])
        >>> dataset = dc.data.NumpyDataset(X=smiles)
        >>> atom_vocab = GroverAtomVocabularyBuilder()
        >>> atom_vocab.build(dataset)
        >>> vocab_labels = GroverModel.atom_vocab_random_mask(atom_vocab, smiles)
        """
        vocab_label = []
        percent = 0.15
        for smi in smiles:
            mol = Chem.MolFromSmiles(smi)
            mlabel = [0] * mol.GetNumAtoms()
            n_mask = math.ceil(mol.GetNumAtoms() * percent)
            perm = np.random.permutation(mol.GetNumAtoms())[:n_mask]
            for p in perm:
                atom = mol.GetAtomWithIdx(int(p))
                mlabel[p] = atom_vocab.stoi.get(
                    GroverAtomVocabularyBuilder.atom_to_vocab(mol, atom),
                    atom_vocab.other_index)

            vocab_label.extend(mlabel)
        return vocab_label

    @staticmethod
    def bond_vocab_random_mask(bond_vocab: GroverBondVocabularyBuilder,
                               smiles: List[str]) -> List[int]:
        """Random masking of bond labels from bond vocabulary

        For every bond in the list of SMILES string, the algorithm fetches the bond
        context (vocab label) from the vocabulary provided and returns the vocabulary
        labels with a random masking (probability of masking = 0.15).

        Parameters
        ----------
        bond_vocab: GroverBondVocabularyBuilder
            bond vocabulary
        smiles: List[str]
            a list of smiles string

        Returns
        -------
        vocab_label: List[int]
            bond vocab label with random masking

        Example
        -------
        >>> import deepchem as dc
        >>> from deepchem.models.torch_models.grover import GroverModel
        >>> from deepchem.feat.vocabulary_builders import GroverBondVocabularyBuilder
        >>> smiles = np.array(['CC', 'CCC'])
        >>> dataset = dc.data.NumpyDataset(X=smiles)
        >>> bond_vocab = GroverBondVocabularyBuilder()
        >>> bond_vocab.build(dataset)
        >>> vocab_labels = GroverModel.bond_vocab_random_mask(bond_vocab, smiles)
        """
        vocab_label = []
        percent = 0.15
        for smi in smiles:
            mol = Chem.MolFromSmiles(smi)
            nm_atoms = mol.GetNumAtoms()
            nm_bonds = mol.GetNumBonds()
            mlabel = []
            n_mask = math.ceil(nm_bonds * percent)
            perm = np.random.permutation(nm_bonds)[:n_mask]
            virtual_bond_id = 0
            for a1 in range(nm_atoms):
                for a2 in range(a1 + 1, nm_atoms):
                    bond = mol.GetBondBetweenAtoms(a1, a2)

                    if bond is None:
                        continue
                    if virtual_bond_id in perm:
                        label = bond_vocab.stoi.get(
                            GroverBondVocabularyBuilder.bond_to_vocab(
                                mol, bond), bond_vocab.other_index)
                        mlabel.extend([label])
                    else:
                        mlabel.extend([0])

                    virtual_bond_id += 1
            vocab_label.extend(mlabel)
        return vocab_label

    def restore(  # type: ignore
            self,
            checkpoint: Optional[str] = None,
            model_dir: Optional[str] = None) -> None:  # type: ignore
        """Reload the values of all variables from a checkpoint file.

        Parameters
        ----------
        checkpoint: str
            the path to the checkpoint file to load.  If this is None, the most recent
            checkpoint will be chosen automatically.  Call get_checkpoints() to get a
            list of all available checkpoints.
        model_dir: str, default None
            Directory to restore checkpoint from. If None, use self.model_dir.  If
            checkpoint is not None, this is ignored.
        """
        # FIXME I am rewriting restore because the restore method in parent class
        # does not restore layers which are not components. This restore method
        # can restore an full model.
        self._ensure_built()
        if checkpoint is None:
            checkpoints = sorted(self.get_checkpoints(model_dir))
            if len(checkpoints) == 0:
                raise ValueError('No checkpoint found')
            checkpoint = checkpoints[0]
        data = torch.load(checkpoint)
        self.model.load_state_dict(data['model'])
        self._pytorch_optimizer.load_state_dict(data['optimizer_state_dict'])
        self._global_step = data['global_step']



"""
Contains class for gaussian process hyperparameter optimizations.
"""
import os
import logging
import tempfile
from typing import Dict, List, Optional, Tuple, Union, Any, Callable

from deepchem.data import Dataset
from deepchem.trans import Transformer
from deepchem.models import Model
from deepchem.metrics import Metric
from deepchem.hyper.base_classes import HyperparamOpt
from deepchem.hyper.base_classes import _convert_hyperparam_dict_to_filename

logger = logging.getLogger(__name__)


def compute_parameter_range(
    params_dict: Dict,
    search_range: Union[int, float,
                        Dict]) -> Dict[str, Tuple[str, List[float]]]:
    """Convenience Function to compute parameter search space.

    Parameters
    ----------
    params_dict: Dict
        Dictionary mapping strings to Ints/Floats. An explicit list of
        parameters is computed with `search_range`. The optimization range
        computed is specified in the documentation for `search_range`
        below.
    search_range: int/float/Dict (default 4)
        The `search_range` specifies the range of parameter values to
        search for. If `search_range` is an int/float, it is used as the
        global search range for parameters. This creates a search
        problem on the following space:

        optimization on [initial value / search_range,
                        initial value * search_range]

        If `search_range` is a dict, it must contain the same keys as
        for `params_dict`. In this case, `search_range` specifies a
        per-parameter search range. This is useful in case some
        parameters have a larger natural range than others. For a given
        hyperparameter `hp` this would create the following search
        range:

        optimization on hp on [initial value[hp] / search_range[hp],
                              initial value[hp] * search_range[hp]]

    Returns
    -------
    param_range: Dict
        Dictionary mapping hyperparameter names to tuples. Each tuple is
        of form `(value_type, value_range)` where `value_type` is a string
        that is either "int" or "cont" and `value_range` is a list of two
        elements of the form `[low, hi]`. This format is expected by
        pyGPGO which `GaussianProcessHyperparamOpt` uses to perform
        optimization.
    """
    # Range of optimization
    param_range = {}
    if isinstance(search_range, dict):
        if sorted(params_dict.keys()) != sorted(search_range.keys()):
            raise ValueError(
                "If search_range is provided as a dictionary, it must have the same keys as params_dict."
            )
    elif (not isinstance(search_range, int)) and (not isinstance(
            search_range, float)):
        raise ValueError("search_range must be a dict or int or float.")
    for hp, value in params_dict.items():
        if isinstance(search_range, dict):
            hp_search_range = search_range[hp]
        else:
            # We know from guard above that this is an int/float
            hp_search_range = search_range
        if isinstance(value, int):
            value_range = [value // hp_search_range, value * hp_search_range]
            param_range[hp] = ("int", value_range)
        elif isinstance(value, float):
            value_range = [value / hp_search_range, value * hp_search_range]
            param_range[hp] = ("cont", value_range)
    return param_range


class GaussianProcessHyperparamOpt(HyperparamOpt):
    """
    Gaussian Process Global Optimization(GPGO)

    This class uses Gaussian Process optimization to select
    hyperparameters. Underneath the hood it uses pyGPGO to optimize
    models. If you don't have pyGPGO installed, you won't be able to use
    this class.

    Note that `params_dict` has a different semantics than for
    `GridHyperparamOpt`. `param_dict[hp]` must be an int/float and is
    used as the center of a search range.

    Examples
    --------
    This example shows the type of constructor function expected.

    >>> import deepchem as dc
    >>> optimizer = dc.hyper.GaussianProcessHyperparamOpt(lambda **p: dc.models.GraphConvModel(n_tasks=1, **p))

    Here's a more sophisticated example that shows how to optimize only
    some parameters of a model. In this case, we have some parameters we
    want to optimize, and others which we don't. To handle this type of
    search, we create a `model_builder` which hard codes some arguments
    (in this case, `n_tasks` and `n_features` which are properties of a
    dataset and not hyperparameters to search over.)

    >>> import numpy as np
    >>> from sklearn.ensemble import RandomForestRegressor as RF
    >>> def model_builder(**model_params):
    ...   n_estimators = model_params['n_estimators']
    ...   min_samples_split = model_params['min_samples_split']
    ...   rf_model = RF(n_estimators=n_estimators, min_samples_split=min_samples_split)
    ...   rf_model = RF(n_estimators=n_estimators)
    ...   return dc.models.SklearnModel(rf_model)
    >>> optimizer = dc.hyper.GaussianProcessHyperparamOpt(model_builder)
    >>> params_dict = {"n_estimators":100, "min_samples_split":2}
    >>> train_dataset = dc.data.NumpyDataset(X=np.random.rand(50, 5),
    ...   y=np.random.rand(50, 1))
    >>> valid_dataset = dc.data.NumpyDataset(X=np.random.rand(20, 5),
    ...   y=np.random.rand(20, 1))
    >>> metric = dc.metrics.Metric(dc.metrics.pearson_r2_score)

    >> best_model, best_hyperparams, all_results =\
    optimizer.hyperparam_search(params_dict, train_dataset, valid_dataset, metric, max_iter=2)
    >> type(best_hyperparams)
    <class 'dict'>

    Parameters
    ----------
    model_builder: constructor function.
        This parameter must be constructor function which returns an
        object which is an instance of `dc.models.Model`. This function
        must accept two arguments, `model_params` of type `dict` and
        `model_dir`, a string specifying a path to a model directory.
    max_iter: int, default 20
        number of optimization trials
    search_range: int/float/Dict (default 4)
        The `search_range` specifies the range of parameter values to
        search for. If `search_range` is an int/float, it is used as the
        global search range for parameters. This creates a search
        problem on the following space:

        optimization on [initial value / search_range,
                        initial value * search_range]

        If `search_range` is a dict, it must contain the same keys as
        for `params_dict`. In this case, `search_range` specifies a
        per-parameter search range. This is useful in case some
        parameters have a larger natural range than others. For a given
        hyperparameter `hp` this would create the following search
        range:

        optimization on hp on [initial value[hp] / search_range[hp],
                              initial value[hp] * search_range[hp]]

    Notes
    -----
    This class requires pyGPGO to be installed.
    """

    def __init__(self,
                 model_builder: Callable[..., Model],
                 max_iter: int = 20,
                 search_range: Union[int, float, Dict] = 4):
        super(GaussianProcessHyperparamOpt,
              self).__init__(model_builder=model_builder)
        self.max_iter = max_iter
        self.search_range = search_range

    def hyperparam_search(
            self,
            params_dict: Dict,
            train_dataset: Dataset,
            valid_dataset: Dataset,
            metric: Metric,
            output_transformers: List[Transformer] = [],
            nb_epoch: int = 10,
            use_max: bool = True,
            logfile: str = 'results.txt',
            logdir: Optional[str] = None,
            **kwargs) -> Tuple[Model, Dict[str, Any], Dict[str, Any]]:
        """Perform hyperparameter search using a gaussian process.

        Parameters
        ----------
        params_dict: Dict
            Maps hyperparameter names (strings) to possible parameter
            values. The semantics of this list are different than for
            `GridHyperparamOpt`. `params_dict[hp]` must map to an int/float,
            which is used as the center of a search with radius
            `search_range` since pyGPGO can only optimize numerical
            hyperparameters.
        train_dataset: Dataset
            dataset used for training
        valid_dataset: Dataset
            dataset used for validation(optimization on valid scores)
        metric: Metric
            metric used for evaluation
        output_transformers: list[Transformer]
            Transformers for evaluation. This argument is needed since
            `train_dataset` and `valid_dataset` may have been transformed
            for learning and need the transform to be inverted before
            the metric can be evaluated on a model.
        nb_epoch: int, (default 10)
            Specifies the number of training epochs during each iteration of optimization.
            Not used by all model types.
        use_max: bool, (default True)
            Specifies whether to maximize or minimize `metric`.
            maximization(True) or minimization(False)
        logdir: str, optional, (default None)
            The directory in which to store created models. If not set, will
            use a temporary directory.
        logfile: str, optional (default `results.txt`)
            Name of logfile to write results to. If specified, this is must
            be a valid file. If not specified, results of hyperparameter
            search will be written to `logdir/results.txt`.


        Returns
        -------
        Tuple[`best_model`, `best_hyperparams`, `all_scores`]
            `(best_model, best_hyperparams, all_scores)` where `best_model` is
            an instance of `dc.model.Model`, `best_hyperparams` is a
            dictionary of parameters, and `all_scores` is a dictionary mapping
            string representations of hyperparameter sets to validation
            scores.
        """
        try:
            from pyGPGO.covfunc import matern32
            from pyGPGO.acquisition import Acquisition
            from pyGPGO.surrogates.GaussianProcess import GaussianProcess
            from pyGPGO.GPGO import GPGO
        except ModuleNotFoundError:
            raise ImportError("This class requires pyGPGO to be installed.")

        # Specify logfile
        log_file = None
        if logfile:
            log_file = logfile
        elif logdir is not None:
            # Make logdir if it doesn't exist.
            if not os.path.exists(logdir):
                os.makedirs(logdir, exist_ok=True)
            log_file = os.path.join(logdir, "results.txt")

        # setup range
        param_range = compute_parameter_range(params_dict, self.search_range)
        param_keys = list(param_range.keys())

        # Stores all results
        all_results: Dict[Any, Any] = {}
        # Store all model references so we don't have to reload
        all_models = {}
        # Stores all model locations
        model_locations = {}

        def _optimize(nb_epoch, **placeholders):
            """Private Optimizing function

            Take in hyper parameter values and number of training epochs.
            Return valid set performances.

            Parameters
            ----------
            nb_epoch: int
                Number of epochs to train model being optimized during each iteration.
                Not used by all model types.
            placeholders: keyword arguments
                Should be various hyperparameters as specified in `param_keys` above.

            Returns:
            --------
            valid_scores: float
                valid set performances
            """
            hyper_parameters = {}
            for hp in param_keys:
                if param_range[hp][0] == "int":
                    # param values are always float in BO, so this line converts float to int
                    # see : https://github.com/josejimenezluna/pyGPGO/issues/10
                    hyper_parameters[hp] = int(placeholders[hp])
                else:
                    hyper_parameters[hp] = float(placeholders[hp])
            logger.info("Running hyperparameter set: %s" %
                        str(hyper_parameters))
            if log_file:
                with open(log_file, 'w+') as f:
                    # Record hyperparameters
                    f.write("Parameters: %s" % str(hyper_parameters))
                    f.write('\n')

            hp_str = _convert_hyperparam_dict_to_filename(hyper_parameters)
            if hp_str in all_results:
                # We have already evaluated the model for these hyperparameters.
                if use_max:
                    return all_results[hp_str]
                else:
                    return -all_results[hp_str]
            if logdir is not None:
                filename = "model%s" % hp_str
                model_dir = os.path.join(logdir, filename)
                logger.info("model_dir is %s" % model_dir)
                try:
                    os.makedirs(model_dir)
                except OSError:
                    if not os.path.isdir(model_dir):
                        logger.info(
                            "Error creating model_dir, using tempfile directory"
                        )
                        model_dir = tempfile.mkdtemp()
            else:
                model_dir = tempfile.mkdtemp()
            # Add it on to the information needed for the constructor
            hyper_parameters["model_dir"] = model_dir
            model = self.model_builder(**hyper_parameters)
            try:
                model.fit(train_dataset, nb_epoch=nb_epoch)
            # Not all models have nb_epoch
            except TypeError:
                model.fit(train_dataset)
            try:
                model.save()
            # Some models autosave
            except NotImplementedError:
                pass

            multitask_scores = model.evaluate(valid_dataset, [metric],
                                              output_transformers)
            score = multitask_scores[metric.name]

            if log_file:
                with open(log_file, 'a') as f:
                    # Record performances
                    f.write("Score: %s" % str(score))
                    f.write('\n')
            # Store all results
            all_results[hp_str] = score
            # Store reference to model
            all_models[hp_str] = model
            model_locations[hp_str] = model_dir
            # GPGO maximize performance by default
            # set performance to its negative value for minimization
            if use_max:
                return score
            else:
                return -score

        # Demarcating internal function for readability
        def optimizing_function(**placeholders):
            """Wrapper function

            Take in hyper parameter values.
            Calls a private optimize function (_optimize) with number of epochs.
            Returns valid set performances.

            Parameters
            ----------
            placeholders: keyword arguments
                Should be various hyperparameters as specified in `param_keys` above.

            Returns:
            --------
            valid_scores: float
                valid set performances
            """
            return _optimize(nb_epoch=nb_epoch, **placeholders)

        # execute GPGO
        cov = matern32()
        gp = GaussianProcess(cov)
        acq = Acquisition(mode='ExpectedImprovement')
        gpgo = GPGO(gp, acq, optimizing_function, param_range)
        logger.info("Max number of iteration: %i" % self.max_iter)
        gpgo.run(max_iter=self.max_iter)

        hp_opt, valid_performance_opt = gpgo.getResult()
        hyper_parameters = {}
        for hp in param_keys:
            if param_range[hp][0] == "int":
                hyper_parameters[hp] = int(hp_opt[hp])
            else:
                # FIXME: Incompatible types in assignment
                hyper_parameters[hp] = float(hp_opt[hp])  # type: ignore
        hp_str = _convert_hyperparam_dict_to_filename(hyper_parameters)

        # Let's fetch the model with the best parameters
        best_model = all_models[hp_str]

        # Compare best model to default hyperparameters
        if log_file:
            with open(log_file, 'a') as f:
                # Record hyperparameters
                f.write("params_dict:")
                f.write(str(params_dict))
                f.write('\n')

        # Return default hyperparameters
        return best_model, hyper_parameters, all_results



"""
SAMPL dataset loader.
"""
import os
import deepchem as dc
from deepchem.molnet.load_function.molnet_loader import TransformerGenerator, _MolnetLoader
from deepchem.data import Dataset
from typing import List, Optional, Tuple, Union

SAMPL_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/SAMPL.csv"
SAMPL_TASKS = ['expt']


class _SAMPLLoader(_MolnetLoader):

    def create_dataset(self) -> Dataset:
        dataset_file = os.path.join(self.data_dir, "SAMPL.csv")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=SAMPL_URL,
                                             dest_dir=self.data_dir)
        loader = dc.data.CSVLoader(tasks=self.tasks,
                                   feature_field="smiles",
                                   featurizer=self.featurizer)
        return loader.create_dataset(dataset_file, shard_size=8192)


def load_sampl(
    featurizer: Union[dc.feat.Featurizer, str] = 'ECFP',
    splitter: Union[dc.splits.Splitter, str, None] = 'scaffold',
    transformers: List[Union[TransformerGenerator, str]] = ['normalization'],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load SAMPL(FreeSolv) dataset

    The Free Solvation Database, FreeSolv(SAMPL), provides experimental and
    calculated hydration free energy of small molecules in water. The calculated
    values are derived from alchemical free energy calculations using molecular
    dynamics simulations. The experimental values are included in the benchmark
    collection.

    Random splitting is recommended for this dataset.

    The raw data csv file contains columns below:

    - "iupac" - IUPAC name of the compound
    - "smiles" - SMILES representation of the molecular structure
    - "expt" - Measured solvation energy (unit: kcal/mol) of the compound,
        used as label
    - "calc" - Calculated solvation energy (unit: kcal/mol) of the compound

    Parameters
    ----------
    featurizer: Featurizer or str
        the featurizer to use for processing the data.  Alternatively you can pass
        one of the names from dc.molnet.featurizers as a shortcut.
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in

    References
    ----------
    .. [1] Mobley, David L., and J. Peter Guthrie. "FreeSolv: a database of
        experimental and calculated hydration free energies, with input files."
        Journal of computer-aided molecular design 28.7 (2014): 711-720.
    """
    loader = _SAMPLLoader(featurizer, splitter, transformers, SAMPL_TASKS,
                          data_dir, save_dir, **kwargs)
    return loader.load_dataset('sampl', reload)
